# -*- coding: utf-8 -*-
""" 
PyCorrFit

Module tools - batch
Stuff that concerns batch processing.
"""


import numpy as np
import os
import wx

from .. import openfile as opf     # How to treat an opened file
from .. import models as mdls


# Menu entry name
MENUINFO = ["B&atch control", "Batch fitting."]

class BatchCtrl(wx.Frame):
    def __init__(self, parent):
        # Parent is main frame
        self.parent = parent
        # Get the window positioning correctly
        pos = self.parent.GetPosition()
        pos = (pos[0]+100, pos[1]+100)
        wx.Frame.__init__(self, parent=parent, title="Batch control",
                 pos=pos, style=wx.DEFAULT_FRAME_STYLE|wx.FRAME_FLOAT_ON_PARENT)
        ## MYID
        # This ID is given by the parent for an instance of this class
        self.MyID = None
        ## Controls
        panel = wx.Panel(self)
        self.panel = panel
        text1 = wx.StaticText(panel, label="Choose source of parameters:")
        self.rbtnhere = wx.RadioButton(panel, -1, 'This session', 
                                        style = wx.RB_GROUP)
        self.rbtnhere.SetValue(True)
        self.rbtnthere = wx.RadioButton(panel, -1, 'Other session')
        self.dropdown = wx.ComboBox(panel, -1, "Current page", (15, 30),
                         wx.DefaultSize, [], wx.CB_DROPDOWN|wx.CB_READONLY)
        # Create the dropdownlist
        self.OnPageChanged()
        text2 = wx.StaticText(panel, label='This will affect all pages'+
                                           '\nwith the same model.'+
                                           '\nApply parameters:')
        btnapply = wx.Button(panel, wx.ID_ANY, 'Apply to applicable pages')
        btnfit = wx.Button(panel, wx.ID_ANY, 'Fit applicable pages')
        # Bindings
        self.Bind(wx.EVT_BUTTON, self.OnApply, btnapply)
        self.Bind(wx.EVT_BUTTON, self.OnFit, btnfit)
        self.Bind(wx.EVT_RADIOBUTTON, self.OnRadioHere, self.rbtnhere)
        self.Bind(wx.EVT_RADIOBUTTON, self.OnRadioThere, self.rbtnthere)
        # self.Bind(wx.EVT_COMBOBOX, self.OnSelect, self.dropdown)
        topSizer = wx.BoxSizer(wx.VERTICAL)
        topSizer.Add(text1)
        topSizer.Add(self.rbtnhere)
        topSizer.Add(self.rbtnthere)
        topSizer.AddSpacer(5)
        topSizer.Add(self.dropdown)
        topSizer.AddSpacer(5)
        topSizer.Add(text2)
        topSizer.AddSpacer(5)
        topSizer.Add(btnapply)
        topSizer.Add(btnfit)
        panel.SetSizer(topSizer)
        topSizer.Fit(self)
        self.SetMinSize(topSizer.GetMinSizeTuple())
        # Check if we even have pages.
        self.OnPageChanged()
        #Icon
        if parent.MainIcon is not None:
            wx.Frame.SetIcon(self, parent.MainIcon)
        self.Show(True)


    def OnApply(self, event):
        # Get the item from the dropdown list
        item = self.dropdown.GetSelection()
        if self.rbtnhere.Value == True:
            # Get parameters from this session
            if item <= 0:
                Page = self.parent.notebook.GetCurrentPage()
            else:
                Page = self.parent.notebook.GetPage(item-1)
            # First apply the parameters of the page
            Page.apply_parameters()          
            # Get all parameters
            Parms = self.parent.PackParameters(Page)
        else:
            # Get Parameters from different session
            Parms = self.YamlParms[item]
        modelid = Parms[1]
        # Set all parameters for all pages
        for i in np.arange(self.parent.notebook.GetPageCount()):
            OtherPage = self.parent.notebook.GetPage(i)
            if OtherPage.corr.fit_model.id == modelid and OtherPage.corr.correlation is not None:
                self.parent.UnpackParameters(Parms, OtherPage)
                OtherPage.PlotAll(trigger="parm_batch")
        # Update all other tools fit the finalize trigger.
        self.parent.OnFNBPageChanged(trigger="parm_finalize")


    def OnClose(self, event=None):
        self.parent.toolmenu.Check(self.MyID, False)
        self.parent.ToolsOpen.__delitem__(self.MyID)
        self.Destroy()


    def OnFit(self, event):
        item = self.dropdown.GetSelection()
        if self.rbtnhere.Value == True:
            if item <= 0:
                Page = self.parent.notebook.GetCurrentPage()
            else:
                Page = self.parent.notebook.GetPage(item-1)
            # Get internal ID
            modelid = Page.corr.fit_model.id
        else:
            # Get external ID
            modelid = self.YamlParms[item][1]
        # Fit all pages with right modelid
        for i in np.arange(self.parent.notebook.GetPageCount()):
            OtherPage = self.parent.notebook.GetPage(i)
            if (OtherPage.corr.fit_model.id == modelid and
                OtherPage.corr.correlation is not None):
                #Fit
                OtherPage.Fit_function(noplots=True,trigger="fit_batch")
        # Update all other tools fit the finalize trigger.
        self.parent.OnFNBPageChanged(trigger="fit_finalize")


    def OnPageChanged(self, Page=None, trigger=None):
        """
            This function is called, when something in the panel
            changes. The variable `trigger` is used to prevent this
            function from being executed to save stall time of the user.
            Forr a list of possible triggers, see the doc string of
            `tools`.
        """
        if self.parent.notebook.GetPageCount() == 0:
            self.panel.Disable()
            return
        else:
            self.panel.Enable()
        # Filter triggers
        if trigger in ["fit_batch", "fit_finalize",
                       "parm_batch", "parm_finalize"]:
            return
        # We need to update the list of Pages in self.dropdown
        if self.rbtnhere.Value == True:
            DDlist = list()
            DDlist.append("Current page")
            for i in np.arange(self.parent.notebook.GetPageCount()):
                aPage = self.parent.notebook.GetPage(i)
                DDlist.append(aPage.counter+aPage.tabtitle.GetValue())
            self.dropdown.SetItems(DDlist)
            self.dropdown.SetSelection(0)


    def OnRadioHere(self, event=None):
        self.OnPageChanged(trigger="view")


    def OnRadioThere(self, event=None):
        # If user clicks on pages in main program, we do not want the list
        # to be changed.
        wc = opf.session_wildcards
        wcstring = "PyCorrFit session (*.pcfs)|*{};*{}".format(
                                                           wc[0], wc[1])
        dlg = wx.FileDialog(self.parent, "Open session file",
                            self.parent.dirname, "", wcstring, wx.FD_OPEN)
        # user cannot do anything until he clicks "OK"
        if dlg.ShowModal() == wx.ID_OK:
            sessionfile = dlg.GetPath()
            self.dirname = os.path.split(sessionfile)[0]
        else:
            self.parent.dirname=dlg.GetDirectory()
            self.rbtnhere.SetValue(True)
            return

        Infodict = opf.LoadSessionData(sessionfile,
                                       parameters_only=True)
        self.YamlParms = Infodict["Parameters"]
        DDlist = list()
        for i in np.arange(len(self.YamlParms)):
            # Rebuild the list
            modelid = self.YamlParms[i][1]
            modelname = mdls.modeldict[modelid][1]
            DDlist.append(self.YamlParms[i][0]+modelname)
        self.dropdown.SetItems(DDlist)
        # Set selection text to first item
        self.dropdown.SetSelection(0)



# -*- coding: utf-8 -*-
u"""PyCorrFit - module "models"

Define all models and set initial parameters.

Each model has a unique ID. This ID is very important:
    1. It is a wxWidgets ID.
    2. It is used in the saving of sessions to identify a model.
It is very important, that model IDs do NOT change in newer versions
of PyCorrFit, because it would not be possible to restore older
PyCorrFit sessions.

Dimensionless representation:
unit of time        : 1 ms
unit of inverse time: 10³ /s
unit of distance    : 100 nm
unit of Diff.coeff  : 10 µm²/s
unit of inverse area: 100 /µm²
unit of inv. volume : 1000 /µm³
"""


# This file is necessary for this folder to become a module that can be 
# imported from within Python/PyCorrFit.

import copy
import numpy as np
import sys
import warnings

## Models
from . import MODEL_classic_gaussian_2D
from . import MODEL_classic_gaussian_3D
from . import MODEL_classic_gaussian_3D2D
from . import MODEL_classic_gaussian_TT3D3D
from . import MODEL_TIRF_gaussian_1C
from . import MODEL_TIRF_gaussian_3D2D
from . import MODEL_TIRF_gaussian_3D3D
from . import MODEL_TIRF_1C
from . import MODEL_TIRF_2D2D
from . import MODEL_TIRF_3D2D
from . import MODEL_TIRF_3D3D
from . import MODEL_TIRF_3D2Dkin_Ries


class Model(object):
    """General class for handling FCS fitting models"""
    def __init__(self, datadict):
        """datadict is an item in Modelarray"""
        self._parameters = datadict["Parameters"]
        self._definitions = datadict["Definitions"]

        if "Supplements" in list(datadict.keys()):
            self._supplements = datadict["Supplements"]
        else:
            self._supplements = lambda x, y: []

        if "Boundaries" in list(datadict.keys()):
            self._boundaries = datadict["Boundaries"]
        else:
            # dummy verification function
            self._boundaries = [[None,None]]*len(self._parameters[1])
        
        if "Constraints" in list(datadict.keys()):
            # sort constraints such that the first value is always
            # larger than the last.
            newcc = []
            for cc in datadict["Constraints"]:
                if cc[0] < cc[2]:
                    if cc[1] == ">":
                        cc = [cc[2], "<", cc[0]]
                    elif cc[1] == "<":
                        cc = [cc[2], ">", cc[0]]
                newcc.append(cc)
            self._constraints = newcc
        else:
            self._constraints = []

    def __call__(self, parameters, tau):
        return self.function(parameters, tau)
    
    def __getitem__(self, key):
        """Emulate old list behavior of models"""
        return self._definitions[key]

    def __repr__(self):
        text = "Model {} - {}".format(
                self.id,
                self.description_short)
        return text

    def apply(self, parameters, tau):
        """ 
        Apply the model with `parameters` and lag
        times `tau`
        """
        return self.function(parameters, tau)

    @property
    def constraints(self):
        """ fitting constraints """
        return copy.copy(self._constraints)

    @property
    def components(self):
        """how many components does this model have"""
        return self._definitions[1]
    
    @property
    def default_values(self):
        """default fitting values"""
        return np.array(self._parameters[1]).copy()
    
    @property
    def default_variables(self):
        """indexes default variable fitting (bool)"""
        return np.array(self._parameters[2]).copy()

    @property
    def description_long(self):
        """long description"""
        return self._definitions[3].__doc__
    
    @property
    def description_short(self):
        """short description"""
        return self._definitions[2]

    @property
    def id(self):
        return self._definitions[0]
    
    @property
    def function(self):
        return self._definitions[3]

    @property
    def func_supplements(self):
        return self._supplements

    @property
    def func_verification(self):
        warnings.warn("`func_verification is deprecated: please do not use it!")
        return lambda x: x
    
    def get_supplementary_parameters(self, values, countrate=None):
        """
        Compute additional information for the model
        
        Parameters
        ----------
        values: list-like of same length as `self.default_values`
            parameters for the model
        countrate: float
            countrate in kHz
        """
        return self.func_supplements(values, countrate)

    def get_supplementary_values(self, values, countrate=None):
        """
        Returns only the values of
        self.get_supplementary_parameters
        
        Parameters
        ----------
        values: list-like of same length as `self.default_values`
            parameters for the model
        countrate: float
            count rate in Hz
        """
        out = list()
        for item in  self.get_supplementary_parameters(values, countrate):
            out.append(item[1])
        return out

    @property
    def name(self):
        return self.description_short

    @property
    def parameters(self):
        return self._parameters

    @property
    def boundaries(self):
        return self._boundaries


def AppendNewModel(Modelarray):
    """ Append a new model from a modelarray. *Modelarray* has to be a list
        whose elements have two items:
        [0] parameters
        [1] some info about the model
        See separate models for more information
    """
    global values
    global valuedict
    global models
    global modeldict
    global supplement
    global boundaries

    for datadict in Modelarray:
        # We can have many models in one model array
        amod = Model(datadict)

        models.append(amod)
        modeldict[amod.id] = amod

        values.append(amod.parameters)
        valuedict[amod.id] = amod.parameters

        # Supplementary Data might be there
        supplement[amod.id] = amod.func_supplements

        # Check functions - check for correct values
        boundaries[amod.id] = amod.boundaries


def GetHumanReadableParms(model, parameters):
    """ From a set of parameters that have internal units e.g. [100 nm],
        Calculate the parameters in human readable units e.g. [nm].
        Uses modeldict from this module.
        *model* - an integer ID of a model
        *parameters* - a list/array of parameters 
                       (all parameters of that model)
        Returns:
        New Units, New Parameters
    """
    stdparms = valuedict[model]
    if len(stdparms) == 5:
        # This means we have extra information on the model
        # Return some human readable stuff
        OldParameters = 1.*np.array(parameters)
        Facors = 1.*np.array(stdparms[4])
        NewParameters = 1.*OldParameters*Facors
        NewUnits = stdparms[3]
        return NewUnits, NewParameters
    else:
        # There is no info about human readable stuff, or it is already human
        # readable.
        return stdparms[0], parameters


def GetHumanReadableParameterDict(model, names, parameters):
    """ From a set of parameters that have internal units e.g. [100 nm],
        Calculate the parameters in human readable units e.g. [nm].
        Uses modeldict from this module.
        In contrast to *GetHumanReadableParms* this function accepts
        single parameter names and does not need the full array of
        parameters.
        *model* - an integer ID of a model
        *name* - the names of the parameters to be translated,
                 order should be same as in
        *parameters* - a list of parameters
        Returns:
        New Units, New Parameters
    """
    stdparms = valuedict[model]
    if len(stdparms) == 5:
        # This means we have extra information on the model
        # Return some human readable stuff
        # Check for list:
        if isinstance(names, basestring):
            names = [names]
            parameters = [parameters]
            retstring = True
        else:
            retstring = False
        # Create new lists
        NewUnits = list()
        NewParameters = list()
        for i in np.arange(len(stdparms[0])):
            for j in np.arange(len(names)):
                if names[j] == stdparms[0][i]:
                    NewUnits.append(stdparms[3][i])
                    NewParameters.append(stdparms[4][i]*parameters[j])
        if retstring == True:
            NewUnits = NewUnits[0]
            NewParameters = NewParameters[0]
        return NewUnits, NewParameters
    else:
        # There is no info about human readable stuff, or it is already human
        # readable.
        return names, parameters


def GetInternalFromHumanReadableParm(model, parameters):
    """ This is the inverse of *GetHumanReadableParms*
        *model* - an integer ID of a model
        *parameters* - a list/array of parameters 
        Returns:
        New Units, New Parameters
    """
    stdparms = valuedict[model]
    if len(stdparms) == 5:
        # This means we have extra information on the model
        # and can convert to internal values
        OldParameters = 1.*np.array(parameters)
        Facors = 1./np.array(stdparms[4])        # inverse
        NewParameters = 1.*OldParameters*Facors
        NewUnits = stdparms[0]
        return NewUnits, NewParameters
    else:
        # There is no info about human readable stuff. The given 
        # parameters have not been converted befor using
        # *GetHumanReadableParms*.
        return stdparms[0], parameters
        

def GetModelType(modelid):
    """ Given a modelid, get the type of model function
        (Confocal, TIR-Conf., TIR-□, User)
    """
    if modelid >= 7000:
        return u"User"
    else:
        shorttype = dict()
        shorttype[u"Confocal (Gaussian)"] = u"Confocal"
        shorttype[u"TIR (Gaussian/Exp.)"] = u"TIR Conf."
        shorttype[u"TIR (□xσ/Exp.)"] = u"TIR □xσ"
        for key in modeltypes.keys():
            mlist = modeltypes[key]
            if mlist.count(modelid) == 1:
                return shorttype[key]
                try:
                    return shorttype[key]
                except:
                    return ""

def GetModelFunctionFromId(modelid):
    return modeldict[modelid][3]
    

def GetModelParametersFromId(modelid):
    return valuedict[modelid][1]


def GetModelFitBoolFromId(modelid):
    return valuedict[modelid][2]

def GetMoreInfo(modelid, Page):
    """ This functino is called by someone who has already calculated
        some stuff or wants to know more about the model he is looking at.
        *modelid* is an ID of a model.
        *Page* is a wx.flatnotebook page.
        Returns:
         More information about a model in form of a list.
    """
    # Background signal average
    bgaverage = None
    # Get the parameters from the current page.
    parms = Page.active_parms[1]
    Info = list()
    corr = Page.corr
    if corr.is_ac:
        if len(corr.traces)==1:
            countrate = corr.traces[0].countrate
        else:
            countrate = None
        ## First import the supplementary parameters of the model
        ## The order is important for plot normalization and session
        ## saving as of version 0.7.8
        # Try to get the dictionary entry of a model
        # Background information
        if len(corr.backgrounds)==1:
            bgaverage = corr.backgrounds[0].countrate
            # Now set the correct countrate
            # We already printed the countrate, so there's no harm done.
        if countrate is not None and bgaverage is not None:
            # might be that there is no countrate.
            relativecountrate = countrate - bgaverage
        else:
            relativecountrate = countrate
        # In case of cross correlation, we don't show this kind of
        # information.
        try:
            # This function should return all important information
            # that can be calculated from the given parameters.
            # We need the relativecountrate to compute the CPP.
            func_info = supplement[modelid]
            data = func_info(parms, relativecountrate)
            for item in data:
                Info.append([item[0], item[1]])
        except KeyError:
            # No information available
            pass
        if countrate is not None:
            # Measurement time
            duration = corr.traces[0].duration/1000
            Info.append(["duration [s]", duration])
            # countrate has to be printed before background.
            # Background might overwrite countrate.
            Info.append(["avg. signal [kHz]", corr.traces[0].countrate])
    else:
        ## Cross correlation curves usually have two traces. Since we
        ## do not know how to compute the cpp, we will pass the argument
        ## "None" as the countrate.
        ## First import the supplementary parameters of the model
        ## The order is important for plot normalization and session
        ## saving as of version 0.7.8
        # Try to get the dictionary entry of a model
        try:
            # This function should return all important information
            # that can be calculated from the given parameters.
            func_info = supplement[modelid]
            data = func_info(parms, None)
            for item in data:
                Info.append([item[0], item[1]])
        except KeyError:
            # No information available
            pass
        if len(corr.traces)==2:
            # Measurement time
            duration = corr.traces[0].duration/1000
            Info.append(["duration [s]", duration])
            # countrate has to be printed before background.
            # Background might overwrite countrate.
            Info.append(["avg. signal A [kHz]", corr.traces[0].countrate])
            Info.append(["avg. signal B [kHz]", corr.traces[1].countrate])

    if len(Info) == 0:
        # If nothing matched until now:
        return None
    else:
        return Info


def GetPositionOfParameter(model, name):
    """ Returns an integer corresponding to the position of the label
        of a parameter in the model function
    """
    stdparms = valuedict[model]
    for i in np.arange(len(stdparms[0])):
        if name == stdparms[0][i]:
            return int(i)
    

# Pack all variables
values = list()
# Also create a dictionary, key is modelid
valuedict = dict()
# Pack all models
models = list()
# Also create a dictinary
modeldict = dict()
# A dictionary for supplementary data:
supplement = dict()
# A dictionary containing model boundaries
boundaries = dict()


# Load all models from the imported "MODEL_*" submodules
for g in list(globals().keys()):
    if g.startswith("MODEL_") and hasattr(globals()[g], "Modelarray"):
        AppendNewModel(globals()[g].Modelarray)

# Create a list for the differentiation between the models
# This should make everything look a little cleaner
modeltypes = dict()
#modeltypes[u"Confocal (Gaussian)"] = [6001, 6002, 6012, 6011, 6031, 6032, 6030]
#modeltypes[u"TIR (Gaussian/Exp.)"] = [6013, 6033, 6034]
#modeltypes[u"TIR (□xσ/Exp.)"] = [6000, 6010, 6022, 6020, 6023, 6021]

modeltypes[u"Confocal (Gaussian)"] = [6011, 6030, 6002, 6031, 6032, 6043]
modeltypes[u"TIR (Gaussian/Exp.)"] = [6014, 6034, 6033]
modeltypes[u"TIR (□xσ/Exp.)"] = [6010, 6023, 6000, 6022, 6020, 6021]
modeltypes[u"User"] = list()




# -*- coding: utf-8 -*-
""" 
This file contains a 3D+3D+T TIR FCS model.
"""
from __future__ import division

import numpy as np
import scipy.special as sps


def wixi(x):
    """ Complex Error Function (Faddeeva/Voigt).
        w(i*x) = exp(x**2) * ( 1-erf(x) )
        This function is called by other functions within this module.
        We are using the scipy.special.wofz module which calculates
        w(z) = exp(-z**2) * ( 1-erf(-iz) )
        z = i*x
    """
    z = x*1j
    wixi = sps.wofz(z)
    # We should have a real solution. Make sure nobody complains about
    # some zero-value imaginary numbers.
    
    return np.real_if_close(wixi)  

# 3D + 3D + T
# model 6034
def CF_Gxyz_3D3DT_gauss(parms, tau):
    u""" Two-component three-dimensional diffusion with a Gaussian
        lateral detection profile and an exponentially decaying profile
        in axial direction, including a triplet component.
        The triplet factor takes into account blinking according to triplet
        states of excited molecules.
        Set *T* or *τ_trip* to 0, if no triplet component is wanted.

        w(i*x) = exp(x²)*erfc(x)
        taud1 = r₀²/(4*D₁)
        κ = 1/d_eva
        x1 = sqrt(D₁*τ)*κ
        gz1 = κ * [ sqrt(D₁*τ/π) + (1 - 2*D₁*τ*κ)/(2*κ) * w(i*x1) ]
        g2D1 = 1 / [ 1+τ/taud1 ]
        particle1 = F₁ * g2D1 * gz1

        taud2 = r₀²/(4*D₂)
        x2 = sqrt(D₂*τ)*κ
        gz2 = κ * [ sqrt(D₂*τ/π) + (1 - 2*D₂*τ*κ)/(2*κ) * w(i*x2) ]
        g2D2 = 1 / [ 1+τ/taud2 ]
        particle2 =  α*(1-F₁) * g2D2 * gz2


        triplet = 1 + T/(1-T)*exp(-τ/τ_trip)
        norm = (1-F₁ + α*F₁)²
        G = 1/n*(particle1 + particle2)*triplet/norm + offset

        *parms* - a list of parameters.
        Parameters (parms[i]):
        [0] n       Effective number of particles in confocal volume
                    (n = n₁+n₂)
        [1] D₁      Diffusion coefficient of species 1
        [2] D₂      Diffusion coefficient of species 2
        [3] F₁      Fraction of molecules of species 1 (n₁ = n*F₁)
                    0 <= F₁ <= 1
        [4] r₀      Lateral extent of the detection volume
        [5] d_eva   Evanescent field depth
        [6] α       Relative molecular brightness of particle
                    2 compared to particle 1 (α = q₂/q₁)
        [7] τ_trip  Characteristic residence time in triplet state
        [8] T       Fraction of particles in triplet (non-fluorescent) state
                    0 <= T < 1
        [9] offset
        *tau* - lag time
    """
    n=parms[0]
    D1=parms[1]
    D2=parms[2]
    F=parms[3]
    r0=parms[4]
    deva=parms[5]
    alpha=parms[6]
    tautrip=parms[7]
    T=parms[8]
    off=parms[9]

    kappa = 1/deva


    ### 1st species
    tauD1 = r0**2/(4*D1)
    # 2D gauss component
    g2D1 = 1 / ( (1.+tau/tauD1) )
    # 1d TIR component
    # Axial correlation    
    x1 = np.sqrt(D1*tau)*kappa
    w_ix1 = wixi(x1)
    # Gz = 1/N1D * gz = kappa / Conc.1D * gz
    gz1 = kappa * (np.sqrt(D1*tau/np.pi) -
                   (2*D1*tau*kappa**2 - 1)/(2*kappa) * w_ix1)
    particle1 = F * g2D1 * gz1

    ### 2nd species
    tauD2 = r0**2/(4*D2)
    # 2D gauss component
    g2D2 = 1 / ( (1.+tau/tauD2) )
    # 1d TIR component
    # Axial correlation    
    x2 = np.sqrt(D2*tau)*kappa
    w_ix2 = wixi(x2)
    # Gz = 1/N1D * gz = kappa / Conc.1D * gz
    gz2 = kappa * (np.sqrt(D2*tau/np.pi) -
                   (2*D2*tau*kappa**2 - 1)/(2*kappa) * w_ix2)
    particle2 = alpha**2*(1-F) * g2D2 * gz2

    ### triplet
    if tautrip == 0 or T == 0:
        triplet = 1
    else:
        triplet = 1 + T/(1-T) * np.exp(-tau/tautrip)

    ### Norm
    norm = (F + alpha*(1-F))**2

    ### Correlation function
    G = 1/n*(particle1 + particle2)*triplet/norm
    return G + off


def get_boundaries(parms):
    # strictly positive
    boundaries = [[0, np.inf]]*len(parms)
    # F
    boundaries[3] = [0, .9999999999999]
    # T
    boundaries[8] = [0, .9999999999999]
    # offset
    boundaries[-1] = [-np.inf, np.inf]
    return boundaries


def MoreInfo(parms, countrate=None):
    u"""Supplementary parameters:
        Effective number of particle species 1:
        [10] n₁ = n*F₁
        Effective number of particle species 2:
        [11] n₂ = n*(1-F₁)
        Value of the correlation function at lag time zero:
        [12] G(0)
        Effective measurement volume:
        [13] V_eff [al] = π * r₀² * d_eva
        Concentration of particle species 1:
        [14] C₁ [nM] = n₁/V_eff
        Concentration of particle species 2:
        [15] C₂ [nM] = n₂/V_eff
    """
    # We can only give you the effective particle number
    n=parms[0]
    #D1=parms[1]
    #D2=parms[2]
    F=parms[3]
    r0=parms[4]
    deva=parms[5]
    #alpha=parms[6]

    Info=list()
    # The enumeration of these parameters is very important for
    # plotting the normalized curve. Countrate must come out last!
    Info.append([u"n\u2081", n*F])
    Info.append([u"n\u2082", n*(1.-F)])
    # Detection area:
    Veff = np.pi * r0**2 * deva
    C1 = n*F / Veff
    C2 = n*(1-F) / Veff
    # Correlation function at tau = 0
    G_0 = CF_Gxyz_3D3DT_gauss(parms, 0)
    Info.append(["G(0)", G_0])
    Info.append(["V_eff [al]", Veff])
    Info.append(["C"+u"\u2081"+" [nM]", C1 * 10000/6.0221415])
    Info.append(["C"+u"\u2082"+" [nM]", C2 * 10000/6.0221415])

    if countrate is not None:
        # CPP
        cpp = countrate/n
        Info.append(["cpp [kHz]", cpp])
    return Info


# 3D + 3D + T model gauss
m_gauss_3d_3d_t = [6034, "T+3D+3D",
                            "Combined 3D diffusion + triplet w/ TIR",
                            CF_Gxyz_3D3DT_gauss]
labels  = [u"n",
           u"D"+u"\u2081"+u" [10 µm²/s]",
           u"D"+u"\u2082"+u" [10 µm²/s]",
           u"F"+u"\u2081", 
           u"r₀ [100 nm]",
           u"d_eva [100 nm]",
           u"\u03b1"+" (q"+u"\u2082"+"/q"+u"\u2081"+")", 
           u"τ_trip [ms]",
           u"T",
           u"offset"
                ]
values = [ 
                25,      # n
                0.9,     # D1
                25.0,    # D2
                0.5,     # F1
                9.44,    # r0
                1.0,     # deva
                1.0,     # alpha
                0.001,   # tautrip
                0.01,    # T
                0.0      # offset
                ]    
# Human readable stuff
labelshr  = [   u"n",
                u"D"+u"\u2081"+u" [µm²/s]",
                u"D"+u"\u2082"+u" [µm²/s]",
                u"F"+u"\u2081", 
                u"r₀ [nm]",
                u"d_eva [nm]",
                u"\u03b1"+" (q"+u"\u2082"+"/q"+u"\u2081"+")", 
                u"τ_trip [µs]",
                u"T",
                u"offset"
                ]
valueshr = [ 
                1.,      # n
                10.,       # D1
                10.,    # D2
                1.,     # F1
                100.,       # r0
                100.,       # deva
                1.,     # alpha
                1000.,       # tautrip
                1.,       # T
                1.      # offset
                ]   
    
valuestofit = [True, True, True, True, False, False, False, False, False, False]
parms = [labels, values, valuestofit, labelshr, valueshr]


model1 = dict()
model1["Parameters"] = parms
model1["Definitions"] = m_gauss_3d_3d_t
model1["Boundaries"] = get_boundaries(values)
model1["Supplements"] = MoreInfo
model1["Constraints"] = [[2, ">", 1]]

Modelarray = [model1]



# -*- coding: utf-8 -*-
"""
PyCorrFit

Module tools - statistics
Provide the user with tab-separated statistics of their curves.
Values are sorted according to the page number.
"""
from __future__ import division

import codecs
import wx
import wx.lib.plot as plot              # Plotting in wxPython
import wx.lib.scrolledpanel as scrolled
import numpy as np
import re

from .info import InfoClass
from .. import misc
from .. import models as mdls


# Menu entry name
MENUINFO = ["&Statistics view", "Show some session statistics."]

def run_once(f):
    def wrapper(*args, **kwargs):
        if not wrapper.has_run:
            wrapper.has_run = True
            return f(*args, **kwargs)
    wrapper.has_run = False
    return wrapper


class Stat(wx.Frame):
    # This tool is derived from a wx.frame.
    def __init__(self, parent):
        self.MyName="STATISTICS"
        # parent is the main frame of PyCorrFit
        self.boxsizerlist = list()
        self.parent = parent
        # Get the window positioning correctly
        pos = self.parent.GetPosition()
        pos = (pos[0]+100, pos[1]+100)
        wx.Frame.__init__(self, parent=self.parent, title="Statistics",
                 pos=pos, size=(700,600),
                 style=wx.DEFAULT_FRAME_STYLE|wx.FRAME_FLOAT_ON_PARENT)
        ## MYID
        # This ID is given by the parent for an instance of this class
        self.MyID = None
        self.MyName = "STATISTICS"
        # List of parameters that are plotted or not
        self.PlotParms = list(["None", 0])
        # Page - the currently active page of the notebook.
        self.Page = self.parent.notebook.GetCurrentPage()
        # Pagenumbers
        self.PageNumbers =range(1,1+self.parent.notebook.GetPageCount())
        ## Splitter window. left side: checkboxes
        ##                  right side: plot with parameters
        self.sp = wx.SplitterWindow(self, style=wx.SP_3DSASH)
        # This is necessary to prevent "Unsplit" of the SplitterWindow:
        self.sp.SetMinimumPaneSize(1)
        ## Content
        # We will display a dialog that conains all the settings
        # - Which model we want statistics on
        # - What kind of parameters should be printed
        #   (We will get the parameters from the current page)
        #   If on another page, the parameter is not available,
        #   do not make a mess out of it.
        # Then the user presses a button and sees/saves the table
        # with all the info.
        self.panel = scrolled.ScrolledPanel(self.sp)
        self.panel.SetupScrolling(scroll_y=True)
        # Parameter settings.
        if self.parent.notebook.GetPageCount() != 0:
            self.InfoClass = InfoClass(CurPage=self.Page)
        else:
            self.panel.Disable()
        # A dropdown menu for the source Page:
        text = wx.StaticText(self.panel, 
                    label="Create a table with all the selected\n"+
                          "variables below from pages with the\n"+
                          "same model as the current page.")
        ## Page selection as in average tool
        Pagetext = wx.StaticText(self.panel, 
                             label="Curves ")
        Psize = text.GetSize()[0]/2
        self.WXTextPages = wx.TextCtrl(self.panel, value="",
                                       size=(Psize,-1))
        # Set number of pages
        pagenumlist = list()
        for i in np.arange(self.parent.notebook.GetPageCount()):
            Page = self.parent.notebook.GetPage(i)
            pagenumlist.append(int(filter(lambda x: x.isdigit(),
                                                        Page.counter)))
        valstring=misc.parsePagenum2String(pagenumlist)
        self.WXTextPages.SetValue(valstring)
        ## Plot parameter dropdown box
        self.PlotParms = self.GetListOfPlottableParms()
        Parmlist = self.PlotParms
        DDtext = wx.StaticText(self.panel, label="Plot parameter ")
        self.WXDropdown = wx.ComboBox(self.panel, -1, "", 
                                size=(Psize,-1), choices=Parmlist,
                                style=wx.CB_DROPDOWN|wx.CB_READONLY)
        self.Bind(wx.EVT_COMBOBOX, self.OnDropDown, self.WXDropdown)
        self.Bind(wx.EVT_TEXT, self.OnDropDown, self.WXTextPages)
        self.WXDropdown.SetSelection(0)
        ## Show Average and SD
        textavg = wx.StaticText(self.panel, label="Average ")
        textsd = wx.StaticText(self.panel, label="Standard deviation ")
        self.WXavg = wx.TextCtrl(self.panel, size=(Psize,-1))
        self.WXsd = wx.TextCtrl(self.panel, size=(Psize,-1))
        self.WXavg.SetEditable(False)
        self.WXsd.SetEditable(False)
        # Create space for parameters
        self.box = wx.StaticBox(self.panel, label="Export parameters")
        self.masterboxsizer = wx.StaticBoxSizer(self.box, wx.VERTICAL)
        self.masterboxsizer.Add(text)
        self.boxsizer = wx.BoxSizer(wx.HORIZONTAL)
        self.masterboxsizer.Add(self.boxsizer)
        self.Checkboxes = list()
        self.Checklabels = list()
        if self.parent.notebook.GetPageCount() != 0:
            self.OnChooseValues()
        self.btnSave = wx.Button(self.panel, wx.ID_ANY, 'Save')
        self.Bind(wx.EVT_BUTTON, self.OnSaveTable, self.btnSave)
        # Add elements to sizer
        self.topSizer = wx.BoxSizer(wx.VERTICAL)
        GridAll = wx.FlexGridSizer(rows=4, cols=2, vgap=5, hgap=5)
        GridAll.Add(Pagetext)
        GridAll.Add(self.WXTextPages)
        GridAll.Add(DDtext)
        GridAll.Add(self.WXDropdown)
        GridAll.Add(textavg)
        GridAll.Add(self.WXavg)
        GridAll.Add(textsd)
        GridAll.Add(self.WXsd)
        #Psizer = wx.BoxSizer(wx.HORIZONTAL)
        #Psizer.Add(Pagetext)
        #Psizer.Add(self.WXTextPages)
        #DDsizer = wx.BoxSizer(wx.HORIZONTAL)
        #DDsizer.Add(DDtext)
        #DDsizer.Add(self.WXDropdown)
        #self.topSizer.Add(Psizer)
        #self.topSizer.Add(DDsizer)
        self.topSizer.Add(GridAll)
        self.topSizer.Add(self.masterboxsizer)
        self.topSizer.Add(self.btnSave)
        # Set size of window
        self.panel.SetSizer(self.topSizer)
        self.topSizer.Fit(self.panel)
        px = self.topSizer.GetMinSizeTuple()[0]
        
        ## Plotting panel
        self.canvas = plot.PlotCanvas(self.sp)
        self.canvas.SetEnableZoom(True)
        self.sp.SplitVertically(self.panel, self.canvas, px+5)
        ## Icon
        if parent.MainIcon is not None:
            wx.Frame.SetIcon(self, parent.MainIcon)
        self.Show(True)
        self.OnDropDown()


    def GetListOfAllParameters(self, e=None, return_std_checked=False,
                                page=None):
        """ Returns sorted list of parameters.
            If return_std_checked is True, then a second list with
            standart checked parameters is returned.
        """
        if page is None:
            page = self.Page
        self.InfoClass.CurPage = page
        # Now that we know our Page, we may change the available
        # parameter options.
        Infodict = self.InfoClass.GetCurInfo()
        # We want to sort the information and have some prechecked 
        # values in the statistics window afterwards.
        # new iteration
        keys = Infodict.keys()
        parms = list()
        errparms = list()

        for key in keys:
            for item in Infodict[key]:
                if item is not None:
                    if key == "fitting" and item[0].startswith("Err "):
                        errparms.append(item)
                    elif len(item) == 2:
                        parms.append(item)

        # Separate checkbox for fit errors
        if len(errparms) > 0:
            parms.append(("Fit errors", errparms))
        
        Info = Stat.SortParameters(parms)

        # List of default checked parameters:
        checked = np.zeros(len(Info), dtype=np.bool)
        # Fit parameters
        pbool = page.corr.fit_parameters_variable
        model = mdls.modeldict[page.corr.fit_model.id]
        pname = mdls.GetHumanReadableParms(model.id, model.parameters[1])[0]
        checkadd = np.array(pname)[pbool]
        for ii, p in enumerate(Info):
            if p[0] in checkadd:
                checked[ii] = True
        # A list with additional strings that should be default checked
        # if found somewhere in the data.
        checklist = ["cpp", "duration", "bg rate", "avg.", "Model name"]
        for i in range(len(Info)):
            item = Info[i]
            for checkitem in checklist:
                if item[0].count(checkitem):
                    checked[i] = True
        # A list with strings that should not be checked:
        nochecklist = []
        for i in range(len(Info)):
            item = Info[i]
            for checkitem in nochecklist:
                if item[0].count(checkitem):
                    checked[i] = False
        
        
        if return_std_checked:
            return Info, checked
        else:
            return Info

        
    def GetListOfPlottableParms(self, e=None, return_values=False,
                                page=None):
        """ Returns list of parameters that can be plotted.
            (This means that the values are convertible to floats)
            If return_values is True, then a second list with
            the corresponding values is returned.
        """
        if page is None:
            page = self.Page
        if self.parent.notebook.GetPageCount() != 0:
            Info = self.GetListOfAllParameters(page=page)
            parmlist = list()
            parmvals = list()
            for item in Info:
                if item is not None and len(item) == 2:
                    try:
                        val = float(item[1])
                    except:
                        pass
                    else:
                        # save the key so we can find the parm later
                        parmlist.append(item[0])
                        parmvals.append(val)
        else:
            parmlist = ["<No Pages>"]
            parmvals = [0]
        if return_values:
            return parmlist, parmvals
        else:
            return parmlist


    def GetWantedParameters(self):
        """
        Updates self.SaveInfo with all the information that will be
        saved to the table.
        """
        strFull = self.WXTextPages.GetValue()
        PageNumbers = misc.parseString2Pagenum(self, strFull)
        # Get the wanted parameters from the selection.
        checked = list()
        for i in np.arange(len(self.Checkboxes)):
            if self.Checkboxes[i].IsChecked() == True:
                checked.append(self.Checklabels[i])
        # Collect all the relevant pages
        pages = list()
        for i in np.arange(self.parent.notebook.GetPageCount()):
            Page = self.parent.notebook.GetPage(i)
            if Page.modelid == self.Page.corr.fit_model.id:
                # Only pages with same modelid
                if int(Page.counter.strip("#: ")) in PageNumbers:
                    # Only pages selected in self.WXTextPages
                    pages.append(Page)
        self.InfoClass.Pagelist = pages
        AllInfo = self.InfoClass.GetAllInfo()
        self.SaveInfo = list()
        # Some nasty iteration through the dictionaries.
        # Collect all checked variables.
        pagekeys = AllInfo.keys()
        # If pagenumber is larger than 10,
        # pagekeys.sort will not work, because we have strings
        # Define new compare function
        cmp_func = lambda a,b: cmp(int(a.strip().strip("#")),
                                   int(b.strip().strip("#")))
        pagekeys.sort(cmp=cmp_func)
        #for Info in pagekeys:
        #    pageinfo = list()
        #    for item in AllInfo[Info]:
        #        for subitem in AllInfo[Info][item]:
        #            if len(subitem) == 2:
        #                for label in checked:
        #                    if label == subitem[0]:
        #                        pageinfo.append(subitem)
        #
        # We want to replace the above iteration with an iteration that
        # covers missing values. This means checking for
        #    "label == subitem[0]"
        # and iteration over AllInfo with that consition.
        for ii in pagekeys:
            pageinfo = list()
            for label in checked:
                label_in_there = False
                for item in AllInfo[ii]:
                    for subitem in AllInfo[ii][item]:
                        if subitem is not None and len(subitem) == 2:
                            if label == subitem[0]:
                                label_in_there = True
                                pageinfo.append(subitem)
                            elif label == "Fit errors" and subitem[0].startswith("Err "):
                                label_in_there = True
                                pageinfo.append(subitem)
                if label_in_there == False:
                    # No data available
                    pageinfo.append([label, "NaN"])
            self.SaveInfo.append(pageinfo)


    def OnCheckboxChecked(self, e="restore"):
        """
            Write boolean data of checked checkboxes to Page variable
            *StatisticsCheckboxes*. If e=="restore", then we will 
            attempt to get the info back from the page.
        """
        # What happens if a checkbox has been checked?
        # We write the data to the Page
        # (it will not be saved in the session).
        if e=="restore":
            checklist = self.Page.StatisticsCheckboxes
            if checklist is not None:
                if len(checklist) <= len(self.Checkboxes):
                    for i in np.arange(len(checklist)):
                        self.Checkboxes[i].SetValue(checklist[i])
        else:
            checklist = list()
            for cb in self.Checkboxes:
                checklist.append(cb.GetValue())
                self.Page.StatisticsCheckboxes = checklist


    def OnChooseValues(self, event=None):
        """
            Plot the values for the parameter selected by the user.
        """
        Info, checked = self.GetListOfAllParameters(
                                                return_std_checked=True)
        #headcounter = 0
        #headlen = len(head)
        # We will sort the checkboxes in more than one column if there
        # are more than *maxitemsincolumn*
        maxitemsincolumn = np.float(19)
        Sizernumber = int(np.ceil(len(Info)/maxitemsincolumn))
        self.boxsizerlist = list()
        for i in np.arange(Sizernumber):
            self.boxsizerlist.append(wx.BoxSizer(wx.VERTICAL))
        # Start at -1 so the indexes will start at 0 (see below).
        #itemcount = -1
        for i in range(len(Info)):
            #itemcount += 1
            #headcounter += 1
            checkbox = wx.CheckBox(self.panel, label=Info[i][0])
            #if headcounter <= headlen:
            #    checkbox.SetValue(True)
            # Additionally default checked items
            #for checkitem in checklist:
            #    if item[0].count(checkitem):
            #        checkbox.SetValue(True)
            checkbox.SetValue(checked[i])
            # Add checkbox to column sizers
            sizern = int(np.floor(i/maxitemsincolumn))
            self.boxsizerlist[sizern].Add(checkbox)
            self.Checkboxes.append(checkbox)
            self.Checklabels.append(Info[i][0])
            self.Bind(wx.EVT_CHECKBOX, self.OnCheckboxChecked, checkbox)
        # Add sizers to boxsizer
        for sizer in self.boxsizerlist:
            self.boxsizer.Add(sizer)
        self.OnCheckboxChecked("restore")
        self.AllPlotParms = Info


    def OnClose(self, event=None):
        # This is a necessary function for PyCorrFit.
        # Do not change it.
        self.parent.toolmenu.Check(self.MyID, False)
        self.parent.ToolsOpen.__delitem__(self.MyID)
        self.Destroy()


    def OnDropDown(self, e=None):
        """ Plot the parameter selected in WXDropdown
            Uses info stored in self.PlotParms and self.InfoClass
        """
        if self.parent.notebook.GetPageCount() == 0 or self.Page is None:
            self.canvas.Clear()
            return
        
        # Get valid pages
        strFull = self.WXTextPages.GetValue()
        try:
            PageNumbers = misc.parseString2Pagenum(self, strFull,
                                                          nodialog=True)
        except:
            PageNumbers = self.PageNumbers
        else:
            self.PageNumbers = PageNumbers
        
        # Get plot parameters
        DDselid = self.WXDropdown.GetSelection()
        #[label, key] = self.PlotParms[DDselid]
        label = self.PlotParms[DDselid]
        # Get potential pages
        pages = list()
        for i in np.arange(self.parent.notebook.GetPageCount()):
            Page = self.parent.notebook.GetPage(i)
            if Page.corr.fit_model.id == self.Page.corr.fit_model.id:
                # Only pages with same modelid
                if int(Page.counter.strip("#: ")) in PageNumbers:
                    # Only pages selected in self.WXTextPages
                    pages.append(Page)
        plotcurve = list()
        for page in pages:
            pllabel, pldata = self.GetListOfPlottableParms(page=page,
                                                     return_values=True)
            # Get the labels and make a plot of the parameters
            if len(pllabel)-1 >= DDselid and pllabel[DDselid] == label:
                x = int(page.counter.strip("#: "))
                y = pldata[DDselid]
                plotcurve.append([x,y])
            else:
                # try to get the label by searching for the first
                # instance
                for k in range(len(pllabel)):
                    if pllabel[k] == label:
                        x = int(page.counter.strip("#: "))
                        y = pldata[k]
                        plotcurve.append([x,y])
        # Prepare plotting
        self.canvas.Clear()
        linesig = plot.PolyMarker(plotcurve, size=1.5, marker='circle')
        plotlist = [linesig]
        # average line
        try:
            avg = np.average(np.array(plotcurve)[:,1])
            maxpage =  int(np.max(np.array(plotcurve)[:,0]) +1)
            minpage =  int(np.min(np.array(plotcurve)[:,0]) -1)
        except:
            minpage = 0
            maxpage = 0
            self.WXavg.SetValue("-")
            self.WXsd.SetValue("-")
        else:
            # Plot data
            plotavg = [[0.5, avg], [maxpage+.5, avg]]
            #lineclear = plot.PolyLine(plotavg, colour="black")
            lineclear = plot.PolyMarker(plotavg, colour="black")
            plotlist.append(lineclear)
            # Update Text control
            self.WXavg.SetValue(str(avg))
            self.WXsd.SetValue(str(np.std(np.array(plotcurve)[:,1])))
        # Draw
        # This causes a memory leak after this function has been 
        # called several times with the same large data set.
        # This goes away if only linesig OR lineclear are plotted.
        #
        #graphics = plot.PlotGraphics(plotlist, 
        #                             xLabel='page number', 
        #                             yLabel=label)
        graphics = plot.PlotGraphics([linesig], 
                                     xLabel='page number', 
                                     yLabel=label)
        
        # Correctly set x-axis
        minticks = 2
        self.canvas.SetXSpec(max(maxpage-minpage, minticks))
        self.canvas.Draw(graphics, xAxis=(minpage,maxpage))

        
    def OnPageChanged(self, page, trigger=None):
        """
            This function is called, when something in the panel
            changes. The variable `trigger` is used to prevent this
            function from being executed to save stall time of the user.
            Forr a list of possible triggers, see the doc string of
            `tools`.
        """
        # When parent changes
        # This is a necessary function for PyCorrFit.
        # This is stuff that should be done when the active page
        # of the notebook changes.
        # filter unwanted triggers to improve speed
        if trigger in ["parm_batch", "fit_batch", "page_add_batch"]:
            return
        elif trigger in ["tab_init"] and page is not None:
            # Check if we have to replot for a new model
            if self.Page.corr.fit_model.id == page.corr.fit_model.id:
                return
        if (trigger in ["page_add_finalize"] and 
            self.WXTextPages.GetValue() == "1"):
            # We probably imported data with statistics window open
            self.PageNumbers = range(1,
                                  1+self.parent.notebook.GetPageCount())
            setstring = misc.parsePagenum2String(self.PageNumbers)
            self.WXTextPages.SetValue(setstring)
            
        #
        # Prevent this function to be run twice at once:
        #
        oldsize = self.GetSizeTuple()
        if self.WXTextPages.GetValue() == "":
            # Set number of pages
            pagenumlist = list()
            for i in np.arange(self.parent.notebook.GetPageCount()):
                Page = self.parent.notebook.GetPage(i)
                pagenumlist.append(int(filter(lambda x: x.isdigit(),
                                                         Page.counter)))
            valstring=misc.parsePagenum2String(pagenumlist)
            self.WXTextPages.SetValue(valstring)
        DDselection = self.WXDropdown.GetValue()
        self.Page = page
        self.InfoClass = InfoClass(CurPage=self.Page)
        self.PlotParms = self.GetListOfPlottableParms()
        # Make sure the selection stays the same
        DDselid = 0
        for i in range(len(self.PlotParms)):
            if DDselection == self.PlotParms[i]:
                DDselid = i
        Parmlist = self.PlotParms
        self.WXDropdown.SetItems(Parmlist)
        self.WXDropdown.SetSelection(DDselid)
        self.panel.Enable()
        for i in np.arange(len(self.Checkboxes)):
            self.Checkboxes[i].Destroy()
        del self.Checkboxes
            #self.Checklabels[i].Destroy() # those cannot be destroyed.
        for i in np.arange(len(self.boxsizerlist)):
            self.boxsizer.Remove(0)
        self.boxsizer.Layout()
        self.boxsizerlist = list()
        self.Checkboxes = list()
        self.Checklabels = list()
        # Disable if there are no pages left
        if self.parent.notebook.GetPageCount() == 0:
            self.panel.Disable()
            self.canvas.Clear()
            return
        self.OnChooseValues()
        self.boxsizer.Layout()
        self.topSizer.Fit(self)
        (ax, ay) = self.GetSizeTuple()
        (px, py) = self.topSizer.GetMinSizeTuple()
        self.sp.SetSashPosition(px+5)
        self.SetMinSize((px+400, py))
        self.SetSize((np.max([px+400,ax,oldsize[0]]),
                      np.max([py,ay,oldsize[1]])))
        # Replot
        self.OnDropDown()


    def OnSaveTable(self, event=None):
        dirname = self.parent.dirname
        dlg = wx.FileDialog(self.parent, "Choose file to save", dirname,
                "", "Text file (*.txt)|*.txt;*.TXT",
                wx.FD_SAVE|wx.FD_OVERWRITE_PROMPT)
        # user cannot do anything until he clicks "OK"
        if dlg.ShowModal() == wx.ID_OK:
            filename = dlg.GetPath()
            if filename.lower().endswith(".txt") is not True:
                filename = filename+".txt"
            dirname = dlg.GetDirectory()
            openedfile = codecs.open(filename, 'w', encoding="utf-8")
            # Get Parameterlist of all Pages with same model id as
            # Self.Page
            # This creates self.SaveInfo:
            self.GetWantedParameters()
            # Write header
            linestring = u""
            for atuple in self.SaveInfo[0]:
                linestring += u"{}\t".format(atuple[0])
            # remove trailing "\t"
            openedfile.write(u"{}\r\n".format(linestring.strip()))
            # Write data         
            for item in self.SaveInfo:
                linestring = u""
                for btuple in item:
                    linestring += u"{}\t".format(btuple[1])
                openedfile.write(linestring.strip()+u"\r\n")
            openedfile.close()
        else:
            dirname = dlg.GetDirectory()
            dlg.Destroy()
        # Give parent the current dirname
        self.parent.dirname = dirname

    def SetPageNumbers(self, pagestring):
        self.WXTextPages.SetValue(pagestring)

    @staticmethod
    def SortParameters(parms):
        u"""
        Sort a list of tuples according to the first item.
        The sorting convention was met in issue #113:
        
        - at the beginning: avg. countrates and particle numbers
        - fast components go before slow components:
          e.g. [τ_trip, τ_diff]
        - model parameters are sorted logically according to their origin:
          e.g. [T1, τ_trip1], or [F1, τ_diff1], or [T2, τ_trip2]
        - if the parameter ends with a number, then we sort it to the
          logical blocks - includes n1, n2, etc.
        - at end: other fitting parameters and mode information

        fitting parameters
        
            intensities
            n
            tautrip1
            T1
            tautrip2
            T2
            tau1
            F1
            tau2
            F2
            tau3
            F3
            alpha
            SP
        
        non-fitting parameter
        
            model name
            chisquare
            weighted fit
            interval
            measurement time
            ...
            model id
        """
        
        startswith_sort = [
                           u"avg. signal",
                           u"n",
                           u"T",
                           u"τ_trip",
                           u"F",
                           u"C",
                           u"D",
                           u"τ",
                           u"τ_diff",
                           u"alpha",
                           u"SP",
                           ]

        otherparms = list()

        # append to this list all parameters that might be in another model.
        for m in mdls.models:
            for p in mdls.GetHumanReadableParms(m.id, m.parameters[1])[0]:
                exists = False
                for sw in startswith_sort+otherparms:
                    if p.startswith(sw):
                        exists = True
                if not exists:
                    otherparms.append(p)
        
        # sort the other parameters by name
        otherparms.sort()
        # special offsets to distinguish "T" and "Type":
        special_off_start = ["Type", "Fit"]
        
        
        def rate_tuple(item):
            x = item[0]
            return rate(x)
        
        def rate(x):
            """
            rate a parameter for sorting.
            lower values are at the beginning of the list.
            """
            x = x.split("[")[0]
            # start at the top
            r = 0
            
            # BLOCK OFFSET
            try:
                intx = int(x[-1])
            except:
                pass
            else:
                # penalty: belongs to block
                r += 3 + intx
            
            # STARTSWITH PENALTY
            for p in startswith_sort:
                if x.startswith(p):
                    r += 1 + 3*(startswith_sort.index(p))
                    break
        
            # Block offset integer
            non_decimal = re.compile(r'[^\d]+')
            pnum = non_decimal.sub("", x)
            if len(pnum) > 0:
                r += int(pnum)
                
            if x.count("3D"):
                r -= 3
            if x.count("2D"):
                r -= 0
        
            # Other Parameters
            for p in otherparms:
                if p.startswith(x):
                    r += (otherparms.index(p)) + 3*len(startswith_sort)

            # Special offsets
            for p in special_off_start:
                if x.startswith(p):
                    r += 300

            if r==0:
                r = 10000

            return r


        def compare(x,y):
            """
            rates x and y.
            returns -1, 0, 1 required for common list sort
            """
            rx = rate_tuple(x)
            ry = rate_tuple(y)
            
            return rx-ry

        return sorted(parms, cmp=compare)


#----------------------------------------------------------------------
# This file was generated by /usr/bin/img2py
#
from wx.lib.embeddedimage import PyEmbeddedImage

Main = PyEmbeddedImage(
    "iVBORw0KGgoAAAANSUhEUgAAAQAAAAEACAYAAABccqhmAAAABHNCSVQICAgIfAhkiAAAAAlw"
    "SFlzAAAMqAAADKgBt04g1gAAABl0RVh0U29mdHdhcmUAd3d3Lmlua3NjYXBlLm9yZ5vuPBoA"
    "ACAASURBVHic7Z13nFTV2ce/d9r2Xcru0quAFDE2sASUJogGS9TYYosxFozGkMTo6/uuEzXF"
    "RoxYIomx94YGjBRpKioqGKQICNKWBRbYXdg27b5/nLne2WXuzOwyc++dmfP9fOYzs3PP7jwM"
    "9/zOc855nucoqqoikUiyE4fVBkgkEuuQAiCRZDFSACSSLEYKgESSxUgBkEiyGCkAEkkWIwVA"
    "IslipABIJFmMFACJJIuRAiCRZDFSACSSLEYKgESSxUgBkEiyGCkAEkkWIwVAIsliXFYbIJFk"
    "K4pXOR64GmgGvlQr1BdMt0EWBJFIzEfxKicD7wElEW/fpFaoj5pqhxQAicRcFK/iAbYB5a0u"
    "qUC5WqFWm2WLXAOQSMznfA7t/AAKJk/LpQBIJOZzg8H7KmI9wDSkAEgkJqJ4laOA0QaXP1Ar"
    "1P1m2iMFQCIxl+tjXHvcNCvCSAGQSExC8SqFwOUGl3cCs0w0B5ACIJGYyaVAscG1f6gVasBM"
    "Y0AKgERiJkaLf0FgppmGaEgBkEhMQPEqJwHHGFyerVao28y0R0MKgERiDkajP1iw+KchBUAi"
    "STGKVykFfmJweRPwvonmtEAKgESSem4Fcg2uPalWWBePLwVAIkkhilfpANxkcNkHPGWiOYcg"
    "BUAiSS03Y7z196Jaoe4x05jWSAGQSFKE4lWKgFsMLoeAP5loTlSkAEgkqeNGoJPBtVfUCnW9"
    "mcZEQwqARJICFK+SD/za4LIK3GuiOYZIAZBIUsMviJ7zD/CmWqGuNtMYI2RNQAmIQhRuwBPx"
    "rL0OIVar/eFn7bXpcevpguJVcoDfxmhyj1m2xEMKQOajIFahOwEdI56L0Dt6e+4DFV0QmoH9"
    "wL7ws/ba1OIWNuJnQHeDa++qFepKM42JhRSAzKII6Ino5FpH7wg4U/BZCpATfhQBpVHaNNBS"
    "GHYBlYjkl4xE8Spu4LYYTe42y5ZEkAKQ3uQDvYDe4ecOcdo7EP/nzvDDgejIkY/W74EY7UPh"
    "ZzXKzyFEpw6En7XItvzwo0eEDQGECGwDtiJEIdTGf7eduRzoY3DtfbVCXW6mMfGQVYHTixxa"
    "dvjOBu1c6B3dFfFQDNonG00IAq1eR+vofmA7uiDsNsnGpBOu9rsO6GfQZJRaoX5koklxkR6A"
    "/XEgbqghwBFEd+ddiLl8Tvg5XkdXEe65P8oj0OpnB2IxMPLhavVzbvhzI+1xhe2JJIC+ZuAL"
    "2+EO//u0TlMLrA0/TK2PlwRuwbjzL7Rb5wfpAdiZbsBQ4EgOTSRxoi/g5WC8nesD6sKP2vBD"
    "+znZbncu4pCL4vCjJPwoxFiQtJ2F5vDr1jfjToQQrAOakmxvCxSvogC5aoXa2M7fLwM20PKg"
    "j0jGqRXqwvbalyqkANiLEsRIP5RD5/MuIA/R0aJ5ASFgD1CFmFfXAu26mZOMA7FI2BnoihC2"
    "AoO2PoTNTbQUgxCwGViDSJ9t9yKi4lWKgcnA2LAt3cJ2dUEIaj1CeKrCjy3AbGBprJJdild5"
    "HOOCn0vUCvW09tqcSqQA2IMewIlA31bvO9A7vTvK7+1F3KQ7EXPndNmbL0IXg64c6uGoCBFo"
    "4tCtxCZgJfAlCXoFilfpDZwdfowh+ncZjxpgDvAuMEetUOsi/v4w4CuiC7MKjFQr1M/b8Zkp"
    "RwqAtfRFdPzIVXJtey2PQ+fQfsSItB3R8X2pN9EUOiDEoC9Q1upaCOEVNNJS4PyITvcFYtQ+"
    "BMWr9EVsu11GchdADwD3AdPVCrVe8SrvAxMN2j6rVqhXJvGzk4oUAGsYgOj4XSLecyJc4zxa"
    "3qwqsAPh+m4jg/fQwxQB/cOPolbXAojOHjlFCABfA8sRHRPFq3QG/geYSsvFyWRTBbyOcb5/"
    "AzBIrVB3pNCGw0IKgHkowGBgJC2371zoHT+SvcC3wHekeAHMxpQjhKAvLTtyECEEjehCEALW"
    "dHuw25Cqg1VejBfjzMSrVqh3WW1ELKQAmEMvYDwtU0NdiBXyyPlvAPgGsZpch0TDiYhwHELL"
    "BJsQQggaGgONyhnPn3Huki1LfmiFgVHYgRj9G6w2JBZSAFJLAXAaYuTX8ITfj5zf+9D3vjNl"
    "Xp8qugDDiYi131633X36c6dPWVe97ohE/4iiusivHYanqRvu5lLczaU4A0UEPPvx5+zBn7OH"
    "5vytNBZubK+dV6oV6rPt/WWzkIFAqcGBqAF/Crrr6kTsj0d2/CbE1tY3iEUtSXx2hR+lwPCv"
    "qr4aOun5ST/ZVb/LqPDG9ziCeRRXn0JJ9akUV5+CM1AY98N8uVXUli2mtmwJBzusBCWh8In9"
    "wIuJNLQa6QEknx4Id19LjlEQI37k3dYArAbWk/mLeiljRdWK3DFPj3mxrrluYKx2SshN6fbz"
    "6fLdlbj88dIljGkq/JbKI56grvTDRJr/Ta1QjcqB2QYpAMkjHzgVEcSjkYMY9bX9YR9iD3s9"
    "mZUAYzq+oE/p8VCPB6obqo223wDouOt0um2ciqepS6xmbaK+ZBU7Bk2noXhtvKbXqhXqP5L2"
    "wSlACkBy6A2ciRABiO7ub0LsWdshOi/tOXLGkdev37t+qtF1RXXQfcPNlG27KCWfH3L42Db0"
    "HvZ3mRermQ+YoFaoS1NiRBKQAnB4KMDJiD19LX02n5bx7zXAp4h5qyQJnPPyOaPe+eadxzAI"
    "7nEGCujz9d0U7z055bZU9XuKqv7/4NA0hu/ZAwxVK9TqlBvTDuQiYPspQIz6vcI/OxERbVqY"
    "aQARqbYW6e4njZqmGsfcb+f+BoPO7wjmMuDLR8k7cKQp9nTd/DOcwXx2DHzYqEkZ8H+I8wFs"
    "hywK2j56Iwo/aJ0/B7Hop3X+rcDbiIU+2fmTyBkvnHFOU6DJYLtPofea/zWt82uUbb2YTpU/"
    "itXkesWrDDDLnrYgBaBtKIitvfMRrr6CCFftGH4dBJYBixAr/ZIksmr3qpwvd35pOO/vuvlq"
    "OuweZ6ZJ39Prm99RUPMDo8tu4I8mmpMwUgASx4Po+CchOrsTEdmnpbbWItJGN1hiXRZwyeuX"
    "XOYP+qMu5+cdHEjXzdeYbdL3KCE3fVZXoIQMEw0vVLzKCDNtSgQpAImRjzjeuXf451xauvzf"
    "Ijp/jfmmZQ8b92883+ha9w1TQbX2dvY0daN0+wWxmlxhli2JIgUgPiXAxegx6EWIxT7N5f8o"
    "/EiXXPy05HfzfjegOdDcO9q1on0jKdp3otkmRaXrd1fFijA810xbEkEKQGzKgUvQO3wJustf"
    "A/wbMfpLUsw7698Zb3St66ZrzTQlJk5/cSwvoKfiVU4w0554SAEwpjfC7dcW+zqgp+xWIqrD"
    "1FpjWvaxvXZ71NU9d3MZBbVHmW1OTDrsHhvrsq28ACkA0RkEnIdeYbcjelTfZuADpMtvGk98"
    "/kR5vb9+aLRrJdWjzDYnLnkHBsUKPT7bTFviIQXgUI4GzkI/OKMzekbfOmApcm/fVOZvmm9U"
    "apviPaPNNCVhSvacanRpcLgCsS2QAtCSQYhMPm2brzN6tOQK4DOL7Mpqqg5WGZ2yS35dVMfA"
    "cvLrhhldchP9GDVLkAKg0xtRLjqy8zsRQd7LgFXWmZbd7G/a37pQKCD23l1+O1T+OhR3c8w+"
    "bnRwqOlIARCUI+ZmmtvfKfwcQkT1yeAeCznoOxjVA3D7bDOQHoK7OapmaUgBsBEdgB+jL/h1"
    "Qh/5lyAq8UospMHfELU3xRllLUV6AOlBAS3j+juiz/k/RST1SCQZSzYLQA5i5C9B3+fXVvu/"
    "QlTtkdiAfHf+nmjv+3NsmWIPxLWt0iw74pGtAqAAP0I/hSayes83CAGQ2IRCT2HUI8P9HjsL"
    "QFTN0pACYDEnAn3Cr4vQI/y2Irf6bEfH3I5Re5Pq8BNw2zMYU3oA9qUnoowXiKw+LbZ/FyLI"
    "R9ZIsxldC7tG9QAAGorXmGlKwjQUrza65Ads47pkmwDkIcp4aXv92ibyAWAhskS3LZnQf8Jm"
    "o2t1Zfast1lbtsTo0jq1wj6FOLNNACajF+zUMvxCwGLkiTy25foTrt9d4C6IOtTXJlaj31Qa"
    "i9bjyzWsAfuOmbbEI5sEYCTikEkQIqAV81gO7LPCIEni9Czp+UG09/05e6gv+dpsc2JSU74w"
    "1uW3zbIjEbJFAHoA2qGROejz/i2IVX+JzTl70NkLjK5V9Z9ppikxCbrrqO75utHl7WqF+rmZ"
    "9sQjGwQgF33e76DlvP9jq4yStI37Tr9vY44rJ2pg1oFOn3Gg06dmmxSVqr5PE3QdNLpsq9Ef"
    "skMARiG2+kDM+7UY/8XIAznTioGdBr5mdK1y4KOJHtyZMny5O2ON/irwLxPNSYhMF4CuiKOk"
    "Qbj9WqTfZ8h5f9rx1kVvvehxeqqiXWss3EBVv3+abdL3qA4/W4Z5UR2GY8oraoX6pZk2JUIm"
    "C4CCntvvQD+ddwcyzDctGdBpgO/4bsfPMLpe1e9f1JRHXStMOduOvI/6DoYBpD7gf0w0J2Ey"
    "WQB+AGh1mYrQq/jaY7IoaRfzLp/3bp4rb2P0qypbh95NY5G567p7er/Mvu7/NrzucrhmqhXq"
    "JhNNSphMFYB89FV/D3qo7yrAcIVGYn8KPAWhSQMm3Y9BxGbI2cTG46ZSV/qRCdaoVPWfyY6B"
    "fzNske/OPzj38rkxjxC2kkwVgNMQ230KItEHoA6w14axpF28ddFbHw/qPOgxo+tBVz2bjv4d"
    "u/u8kDIbQs5GNh99O1X9nsIoetypOIMPTHzgmbF9x/ZHTzyzFZkoAL2AIeHXBbTM75fFPDOE"
    "VTes+ntpfulcwwZKiMoBM9g8/HZ8eTuS+tkHOn3O+hE/p7Zsccx2t55867wbTrjhO8RANCGp"
    "RiQJRbVPWHKyuAJRdNEZflaA7xDVfSQZxIqqFbmjnhr1bIO/YUisdkrITemO8+iy+Wpc/g7t"
    "/rzGwg1UDniUA53jLyNNHjj5izmXzvkAaEQ/P+I/gK2ylzJNAAYCU8KvtVr+AUQAhjytNwN5"
    "bPljXabNnfaPpkBT33htHcE8iqtPoaT6VIqrT4l1hNf3+HJ3UVu2mNqyxRzssDKhWIPS/NK5"
    "227d9mKuK1crDbwPsROwH3gaG2WcZpoA/BRR4NOFXnr5C8AwN1OS/szfNL/wwtcufKCmqeaH"
    "8VsLFNVFfu0wPM3luJpLcfs64wwUEHDX4M/ZS8Czj+b8bTQWtqkerDqw88Anvr7h68c9To8D"
    "cbhMPqLza3En/8ZG29CZJAD9EF846KN/E/AGMs0347nwtQtPfH3N609i0bqWQ3E0jek75n8W"
    "XLEgcl1iMCIJDWAvIvJ0D/Cc2fYZkUmLgNrxsC708l5rkJ0/45m+bHr3t9e9fT8W3c95rryN"
    "N5xwwxWtOj+IcvKN4dfafKMM6G+acXFwxW+SFvREL7WsZfr5kJl+Gc8rq1/pePsHtz8ZCAU6"
    "mv3Zbqd793Hdjpux4PIFswo8BdEWB4KIQeh4xKDkQqxJnQjYIjAoUwRAG/2d6EE/a5HJPhnN"
    "su3L8q+Zdc3jzYHmPkZtclw5W5sDzb0Qu0FJwaE46gd1HvTUqxe++szw8uHNcZp/AxyFnoZe"
    "C3RDnERledn5TBCArugFPrXRP4AQAEmGsqVmi3vy85MfrvfXGx7CV5RT9OW6qeuueW3Na2VP"
    "r3x6zOaazWPrmupGqKhtvu+divNA5/zOS4eUDll015i7lo7pOybRiFLtXjwGMTgdRHgGI7GB"
    "AGTCIuA5wBGI0V+LtvoasF3mlSQ51PvqHX0f7nt/dUP1RKM2Hqdnxws/fuGSC4ZesD/y/UXf"
    "LSr0LvaO2rR/0zEN/oaypkBTuS/oK/eH/GWqqrodiqPe4/Ts8Tg9u3NduXsKPYU7j+9+/LIZ"
    "k2d8UV5Q3t71JA/iABo3LeMCXgJ2tvNvJoV0F4AC4BfoIb/5CHV9A7EDIMlAek7veeeOuh0X"
    "GV13OVz7Kk6ruPzOU+9s0wi7aveqnARc+vZyLHpq+h7Effo1YBzNaALpvgswGNH5FfS5/wZk"
    "589YBs0YdGOszu9UnAevO/6669va+QFS2Pmh5Y6Udq8OwuJpeLoLgHY4vJb4A2CQKipJd45/"
    "8viLN+zdcIPRdUVRfD8e+uNbZpw5w47rP83oB83mhp89wABrzBGkswCUos/5NUWtQVb6yUjG"
    "PTNu0pc7v7w9RpPQuL7jfv/qBa/a+WSnb8PPLvSq1EMN2ppCOguAlgDiQA/8+dagrSSNOefl"
    "c0Yt2rLoT8S4X0/ofsI986+Yb9u8+zCV6NNTbdDqg757ZTrpKgAKugBo7pQKGJ4gI0lPprw0"
    "ZfS76999WFVVt1GbI0uPnLH82uWGBUNtROQ9moe+fhUzmzGVpKsA9EIPrdSUdCcy4y+jOOvF"
    "s06bvWH2X1VV9Ri16VXc68V1U9f93Uy7DhPNS1XQi9RaNg1IVwHQvjAn+lzKFqGVkuQw+YXJ"
    "Y9/b+N70WJ2/rKDsvY03b/yzmXYlgX2ItSoQ29bQcj3LVNJRAFyIvH/QR/8ANoiqkiSHSc9P"
    "Gvf+xvcfiuX2l+SWLFt53co7PE5POgayaF5ADnofHGyFIekoAD3QR31t/r8FIQKSNOf0506f"
    "MO/beQ/GCtctyin68r3L3vtV96Lu6fp/HrlWpS1gG+YzpJJ0zAXoFX52oNu/3SJbMo511ety"
    "/r3+351X71ldVnmgsnO+O7+pf8f+e0b2GFl93uDzalI54o5/dvzEhZsX3qeiOo3aFOcUfz7/"
    "ivk3jug+otGoTRrQgJgKdEKsAzQipgC5mBzElo4C0Dv8nBPxnuFZzBJjfEGfctOcm4Yt2Lxg"
    "fNXBqlHNgeaeQTVoWCdLQQm6nK69RZ6i1YM6D/pg2snTFreOtW8v454ZN2nRd4v+Eqfzf7bo"
    "qkVTj+16bCZEeu5ECIB2HyuItHZTA9nSLRfAA0xFfFkliDWAGmx25rrdufLtK4/5YPMHZ+2q"
    "3zXOH/SXt/fvKCihwpzCFX079P3goYkPvTmh/4R2nbkw5ukxk5dsWfJnFdVwSlqSU/LpwqsW"
    "3pQhnR/EVHZ8+LWWG7ASMPVoo3QTgP7AueHXZYhdgLXAcsssSiN+N+93A/6x4h+/2t+4/7Rk"
    "/22nw1k7tHTok29e9ObLAzoN8CX6e6P/Nfqsj7Z+9MeYnT+3ZNnSq5f+MsWx+mbjAi5BDGa1"
    "iGnAXuAZM41INwE4DVFdJTL1dyF6jLUkCo8tf6zL3UvuvqnqYNXZpHjh1+P0VI7oPmLG+z99"
    "f7ZBlZzvOfrxo69YtXvVb4hRrKNDboePl12z7ObBpYMzqfNrTEbcx03oW4NPYGI8S7oJwOWI"
    "LywPMQUAeBlR/ksShXHPjpu4+LvF94bUUG781smjyFP05RM/euLWS4dfekhuhi/oUwY9Mug3"
    "W2q3XBHrb3TI7fDhsmuW/SpDOz/oKcIhYHf4vdmYWMounbYBc9FHfS04ZC+y80cl3MmmLty8"
    "8AGzOz/AAd+B466edfVL0+ZOGxj5/sZ9Gz09H+p5f7zO3zG349Ll1y6/JYM7P4B21HnkjlZv"
    "g7YpIZ0EoEfEa23lNOpZ8dnO8srleT0e6vHghn0brieJtfDaii/o6/7XT/76/OQXJo8FeH/j"
    "+0XHPXnck3sa9kyK9Xsd8zouXnHdilvaspaQpuxGP65Ou6d7mmlAOk0BRgCjEaKlrVzL+X8r"
    "VlStyB39r9FP1fvqh8dvbRqhYWXDHt60f9OUxkBjzPz30vzSeZ9f+/ltfTr0yZaCrmcg7ufI"
    "UmF/w6TAtnSKA+gUfo60uSZaw2zFF/QpE5+beK/NOj+AY/We1bfGa9SruNeLa6eu/Uu8xcMM"
    "Yz/6aVYaHRFbgyknnQUghKiwKgkz/PHh18UqlGlj1GFlw6Z/fePX/7LaEAuoCz9bIgDptAag"
    "HfygfVEHsNEhi1Yz4dkJp6/fu/5Gq+1oK4qi+Ef3Gf37LO38oAuAgt4fOxm0TTrpIgB56Ik/"
    "WqhorUHbrOPplU+XLvxu4T1YuODXHpyK8+D5Q86/YclVS+ZYbYuF1EW81gY30045ShcBiPxC"
    "tC+pLlrDbOTOD+68IaSG8uO3tA8uh2vvL0/85ZWvXfjap1bbYjEH0XcCtMFNegCt0L4QBekB"
    "tOCuRXf1qTxQeb7VdrSVwpzCtdMnTbfNMdkWoiKmsyA9AEM0AYjMFJMeAPDo8kdvjpVBZ1dq"
    "GmtGXfn2lcdYbYdNaL0Q6EEveZdS0kUAWi8AgvQAuGnOTUPSdNUfgFnrZt1ktQ02wWgnIOWk"
    "iwBoaqiNdD5kCDBzNsyZbLUNh0Ndc92IV1a/Yvqx3jZEE4BITy4vWsNkky4CoMX+a6vc2RIl"
    "FpPKg5XjrLbhcFBRHdM/mT7GajtsQORgpt3jhsVQk4kUgDTl9/N/f0RzoNmSOnLJZMPeDWkt"
    "Ykki8n6WAhAFrQioFIAws76ZNT5+K/uzv2n/ycsrl5vi7toYKQBxaO0BZP38f0fdjh9abUMy"
    "UFU1x7vIO8JqOyxGCkAMImvDa/amaznopNEcbO5utQ3JYufBnRnzb2knkQKg3eNSAMJEfhHS"
    "A0Bk/fmD/lKr7UgWtc21lpyKYyOkBxCDaAKQ1WsAr695vWOsgzPSjXpffbsrE2cIUgBiEDkF"
    "kAIAfLL9k4zqME2Bpmz3AFREWXCQAnAIkV+EZm9WC8CW2i0Z4/4D+IK+jPr3tBNtXUuuAbQi"
    "GPFay/9Pu9j3ZFKSU1JvtQ3JxKk4M+rf0060vqjd40Gjhqn4UDsTueCnfTmGp8ZmA0eVH7U7"
    "fqv0IdeVa0r1G5uj3dNaarApC91SANKQs488u9pqG5JJnjsvowStHUQu6Gr3uBSAMJHzfSkA"
    "wODSwc1OhzNjsiGLPEUZJWjtIPJ+lgLQCukBRMHj8GSM29wxr2O2ewBSAGIQQl8Q0eZHWS8A"
    "JbklX1ttQ7IY32/8KqttsBgpAHHQvgzpAYQ5pusxph4jnSpyXbmb7hpz1xar7bCYyPtZLgJG"
    "obUAmLJHamemT5r+sUNxNFptx+HSo6jHAqttsAHRPABTzkRMFwHQFgKlBxBmcOng5o55HT+y"
    "2o7DZVy/cRnhyRwmcgoQh6bws7YWkEP62J4yhpQOSevR0+1075px5ozVVtthAwrCz5EH3UgB"
    "iGB/+DkyXrrIIltsw/PnPf++x+nZYbUd7eXIzkc+53F65OlOUBx+jkxzN2WbN10EYF/4OfIL"
    "Ko7WMJvo06GPf2SPkY9YbUd78Dg9lW9d9NZLVtthE1oLgIpJB9+miwBoHkAI3U0qscgW2+AL"
    "+pRAKOBRUEyJG08mI7qPmDGg04CsrusQQWsBqMWkXIB0ySnfH/E6gFg0yWoPYNrcaQNnfjHz"
    "fw/4DhxrtS1tJd+dv+79n74/22o7bEJO+AF6p99v0DbppIsA1CBGfwe6AGSlB7B069KCy9+6"
    "/MatNVsvS8cTgRRF8V807KK7CzwFofits4LIgUzzAPZFa5gK0kUAtDlRJ3SVzDoPYNyz4yZ+"
    "uPXD2/xBf9oWBBnZfeQ9T53z1H+ttsNGRN7H0gOIwT6EAGgqqblOpgRMWMkfFv+hz/RPpt9R"
    "01RzitW2HA69S3o//8nPP3nTajtshiYAQfT1LekBREFTxcidgBIgYxNJ1lWvy5ny0pRrv93/"
    "7dWqqiYU/ViUU/TlQd/BoxJtbxYluSXLvrr+q/uttsOGaFPZyPtaegBRaB0LANAZAwH449I/"
    "9pq/ef6Q/Y37S+ua68oa/A2lTYGm0qAazM915u7Oc+dVF3oK95TkluztU9Jnx4MTH1zZvai7"
    "bcqNT3lpyuh53867oznY3DOR9h6np3J8v/F/mnPZnEVTXpoy+r0N790XVIOmnDAbjw65HT58"
    "7cLXftsht4Oc9x+KVg5Nu/eaAdMqJCmqmjZxGKXAFeHXnRD5ANuAhVqDqXOmDpn77dzxO+p2"
    "jG8MNA5oyx93Ks6DnfM7LxlWPmzBvePu/fDknic3JM3yNvDIZ4909S7y/n5v496ETv5RFMXf"
    "v0P/p9+6+K2/Dy8f/v106I4Fd/R/6JOHHmkONPdOnbXx6dOhz7Mrr1v5oOz8USkGzg2/3o/o"
    "/JuBt8wyIJ0EAOAGxKmpheGHb0vtljenvDTl4m+qv/mpL+jrlowPURTFV5ZfNv+XI3/56J2n"
    "3rk1GX8zHpUHKl0Tn5t4xdrqtdeH1FBCR2UV5xR/esuJt9zzh7F/+C7a9dkbZhdf9uZlD9Q2"
    "1Z6cVGMTQFEU/0k9TvJ+fM3Hs8z+7DRiEHBS+PUuxBrAYuALswxINwH4EeJL84TUUKe7Ft01"
    "9IGPHzi+MdDYNRUfpqAEuxd3f/2esfc8cdUxV6Wsas0lb1xywtvr3r6zKdB0RCLtXQ5X9Sm9"
    "Trl/8VWL58RrW9NU4zjpnyddu2HvhqtDaqggXvtkUOgp/Orioy7+y8wpM7M9zz8epwF9EMlu"
    "e8PvPY+J61rpJgBHAxP+/vnf+9+z9J4Lt9dtN2U7zKE4Go/oeMQzi65a9PdkrhM899/nOv92"
    "7m+n7arfNSWR9gpKqFdJr5eeOfeZGWP6jjnYls96cdWLnW6bf9sNO+p2XJiq+IEcV86WUb1G"
    "PTz/ivnzUvH3M5CLEDtZB8OPJuBxWiYFpZR0E4CO17577cynVjx1XkgNmR7GXJxT/Nlz5z33"
    "67OPPPuwEjXqffWOU58+9SdfVX11c1ANJpTUVOgp/O+VP7jynhlnzlh7OJ9916K7+jy2/LFf"
    "VjdWj03WToHH6akcWjb06dmXzn7NTgupNqcjoAn/PkT230bgHTONSBsBULyKC3gYuLEtv+cA"
    "OigKpQ6FMkUhF9irquxRVapDapuDCHJcOVt/c/Jvbrpn3D2b2/irAJz90tmj5m+aPy3RRUqn"
    "w1l7bNdj//rRzz56I5mZc8srl+fdseCOH67avWpsdUP1qcFQsEMbfl0tcBes6Vncc+GkAZMW"
    "PXzGw98ky64sYgignYqszf8XAivMNCItBEDxKh2AN4BxibQvUxRGuZyMdjk5+TyThgAAENxJ"
    "REFU1uWMWT1ke0hlaSDA4kCQ1cFQQr6XU3EePGvQWb+ddfGsDxOxB8QOxfP/fX5aXXPdiQn+"
    "itq1sOusRyY/8tAFQy9I6b5wTVONY9rcaUdt3Lex577GfWUHfAfKG/wNZc2B5jKn4mzMc+ft"
    "LnAX7CnJLdldll+257wh56259rhrMzb+wiTGAr0QI78W+PMM+lqAKdheABSv4gbmAmPitT3a"
    "6eAXOR5+4HR8f8BaW9irqrzqC/Cazx+3GoOiKP7Lj778Z8+c+8zKWO2mL5ve/c8f/fnm3fW7"
    "z4TEzMpz5224YMgFdz973rOmjgYS01AQ838P+vy/AXjCdEPSQAAeB66P1aavw8H1OW5GuZKz"
    "trVbVflns5/3/AFibV67HK69D5/x8EU3jrhxV+trszfMLr5pzk3Xbqndcmmic22H4mgYVjbs"
    "0flXzH+hvKA87VJ8JQnTE92b3YvYBfgGMD1D0tYCoHiVqcCMWG3Odrv4da4nJSGNK4JB7mz0"
    "URvjOypwF6xZ9vNlV2hBOBv3bfT8+JUfX7ymes11wVAw4YSl0vzSuXePvfsv159wvXStMx9t"
    "+y8AaNvLc4B1ZhtiWwFQvMoYYB4G4coO4KYcDz/xpDaauTKkcltjM5tDxr5AeUH5nG23bvv9"
    "hOcmTP50+6e3+IK+7on+/TxX3sZJAybd/9ZFb32cFIMldscD/ARxC2vuvw/h/pu+g2JLAVC8"
    "ihNYhVgpPQQHcG9eDqOT5PLHo0FVubWxmdVBYxHIdeVuaQo09Un0b7qd7t3HdTtuxoLLF8yS"
    "ufFZRWT03x5EbsvXiHUu07FrMtBVGHR+gJtzPKZ1foB8ReFPeTlcU9/EHgPBTLTzOxXnwYGd"
    "Bz716oWvPhsZuy/JGrRoTx96Ytsai2yxX01AxavkAV6j6+e6XVyQYrc/Gp3CItDeyBkFJdCr"
    "uNeLL1/w8plrp66dKTt/VlIElIVfa4e61AHbrTHHnh7ALUCPaBcGOBzcmmtdmvtgp4Obcj08"
    "1NS2Wpal+aVzbznxlofNSiyS2BZt9FfRz7o4rMjOw8VWAqB4FQfwa6PrN+a4sboI3jluF2/4"
    "AmyJsSioUZRT9OWFQy988J9n/1OWwJIA9A8/N6HH+1vm/oPNBAD4IbqL1ILjnU5GmjjvN8IJ"
    "3JDj5veNxh58rit305i+Y/763mXvLTRsJMk2eiNS2EF3/3diYvWfaNhNAM41unB9jn2OAxzl"
    "cjLc6WBVlF0Bh+JoXnHdigsHlw6WNe8lkQwPPwfQj/2y/Fg0uy0CRhWAng6FIU57mTrRHV07"
    "Q2oo5/YFt59gsjkSe9MdUb4OxL4/iNBfS91/sJEAKF7laPQ5UgtOddnNUYHRLqdhYP/yyuUJ"
    "JS1JsobI0V9b/PsCCwJ/WmMbAQBGGl0wc88/UUoVhaEGXklNY83wqBck2Ug50CX8Wiv22QR8"
    "ZY05LbGTAESt55erwDCbuf8axzijC5Mv6Iu6kCnJSo4OPwfRR/8VmHT8dzzs1LOixs+XKYqt"
    "jIykVIk+CQiEAp1rmmrsarbEPDqh39cHEVt/Pkwu+hELO92kUQWgVLGTiS0pdUQXABXV8cba"
    "NzpHvSjJJrTRP4Q++n8V8dpy7NS7onsABp3MDhh5AACfV34upwHZTVfE3j+Iub+KWPQzreR3"
    "IthJAKIWx8w324o2kBdDm/Y37jelBLfEljgArfRbALHlByLD1ZIDZ4ywkwBURXtzrw3TlTX2"
    "hYxtG9BpQMrOEZDYnqHoZ/4dQJ/7L7fMIgPsJACV0d6strEAxLJtQv8Je0w0RWIfCoAfhF83"
    "oZ9evQw9CMg22F4A9sQYZa3GyDtxKI7Gth7cIckYRiJSRlTE6A+i8MeXllkUA9sLwD5Vpc6m"
    "XsB2A3FyO92yrl920hNR6hvEaK8V/JiPiaf9tAU7CcCWaG+GgI8C9iuQGwKWGdiV68zdaa41"
    "EhvgRI9mjVz4+xqR9WdL7CQACxDlkQ9hiQ0FYHUwxD4Dz6RXSa+EDwyRZAzHoqf71qEX/Vhi"
    "mUUJYBsBUCvUGmBRtGufBYI02cyBWhpDlC456pIPTDRFYj09ESv/IHL9tTDfJdgo6CcathGA"
    "MG9He7MZmOW3PHHqew6qKrMN7Mlz5224Y/Qd20w2SWIdBYhCNiBc/7rw60qE+29rbCMAilcp"
    "AwYaXX/W56feJouBz/oChoeF9CjqIUf/7EEBTkUc8Q1Qg77n/x+rjGoLlifaK17leOBm9LPS"
    "o1KrqrzoC3CtxZWBdoZUXvNFXaoAUM8bfF5a/MdLksKx6CXs6tDz++cixMD2WCIA4QM/zwd+"
    "CZyS6O+94vczwe2kn8MaxyUEPNjsi75SCZQXlL933+n3bTTTJoll9ACOCr9uRF/1/wpYb4lF"
    "7cBUAVC8SjlwHeKwz4SPz9JoUuG2xmZm5udSEiMRJ1X8vdnPJwaLf4qi+KedPO1vJpsksYZ8"
    "YFT4deS8fw+w2BKL2okpR4MpXuUExGgf081PlGOcDv6an2uqer3vD3B3jPMAepf0fn7Lr7b8"
    "xUSTJNbgAiYCpeGfqxEi4Aeex+Iqv20lZX0ows2/GTg5mX97ZTDE/zY283+5OTEz8pLFPH+A"
    "P8fo/E7FefDBiQ8+mXpLJBajIE721Tp/Lfq8fx5p1vkhBR6A4lW6AL+gnW4+Yj71AuK01KfR"
    "51mH0N/h4C95OXRLUc2AEDCz2c9zxot+AKEzB5558+xLZ6eV6ydpF6PQC9fWo8f6r0IIQNqR"
    "FAEIn+gzCfg5MAVoz1L9VuAx4B9qhbo3/Hf7AZ9icFgIQImicFuuh1OTXDh0j6pyf5OPj+NE"
    "IQ4tGzp99Y2rn0rqh0vsyAm0DPapDb+uBF7HBhV+28NhCYDiVXoDPws/esVpbsRi4G/ALLVC"
    "PaS3KV5lNCKZIuahgD9wOpia4zGs1JsojSo87/Pzit8fN/qwS0GXf1f9pur2w/pASTpwFHBc"
    "+HUz+n7/XuAVbB7tF4s2C0B4bn82YrSfSPuCiTQ3/xG1Qo17bp7iVa4G/gmGpfi/51SXk9Pd"
    "Lk5yOtu0PrAhFGJpIMjbvoBhjH8khZ7Cr5Zfu/yawaWD5Sm/mc0A9K1qP7APPdX3ZfRpQFqS"
    "sAAoXuVI4BrgSkSt8/awBd3N39eWX1S8ynnAs+gJFzFxAyNcTo5zOihVHHRyiCO+c1HYp6rs"
    "VVX2qSrfhTt+VRvqDpQXlM+Zf8X8/5NHfGc8vRGLfgrCxd+HXuDz5fDPaU1cAVC8ykjgfkTI"
    "Y3tZhHDz34nm5ieK4lWGA7OAfodhy+GgDi4d/MjaqWtnWvT5EvMYCJyE6PwhhLsfRAjBa9g4"
    "xbctxBQAxaucCvwbg4KdcTgAvATMUCvUVe0zL6pNnYE3EMpsGk7F2Tih/4Tb/vPT/8gTfzOf"
    "o4Fjwq9DiJE+EH79DrDJIruSTjwB2EHbt/I+A2YCL6sVakrKYileRQEuBe4F+qTiMzQciiM0"
    "vt/4/04/Y/qiYWXD3kaMBJLMRAFGAIPDPwcQe/ua1/o+NjjRN5nEE4Aa9OqmsahBREHNTGRR"
    "L1koXiUHEWF4B9Ax2X//qPKjVj048cH5E4+YqG3xBBDTmajlyyRpjROxz68NKH5E5w+FH/8B"
    "1lljWuqIJwCVGJzZF+ZDxGj/mlqhNibZtoRRvEpH4CrgHMR/4uEEBWxzKI537x5796Y7Rt+h"
    "fTkehMBo88GPgM2H8RkSe+EGxiIO84CWW31+hNsftWRduhNPAE5ElOqKPORiL/AMYiV/bWrN"
    "azuKV+kEnIXYqjwOMYXJNWgeQiRwbEYo/Dtqhaqd26YAYxApnyDCpjuhb3uuQESASdKbQkTn"
    "1zzIRvSSXo3AWxicWZEJJLILcCIilj8f+BZ4W61Q02r7K+wh9ECIQR5iBbcSqFIr1HgRXCei"
    "V3xxIm4ULYdiB8ILSqvvQ/I9vRD/t1qQ2UH02v11iMXmtIvvbwumZANmAEcBpyO8AgfQAf2m"
    "aUBEM8qDQNIHB8I71EJ7VUSH16ax1cCb2PAgj2QjBSBx+gM/Qh/9C9GDklTEwQ8ZtUKcoRQg"
    "Ylq0/JIAYr6veYI7ELUps8KrkwLQNroh1heKwz97EN6Ati6wHbFAmBU3TxrSA7FIrNWkiJzv"
    "A/wXscuTlok97UEKQNvJQWQ+Dgj/7EBslWo3VT1iXWCX+aZJDHAiAnuGhX9u7fL7EHX80qaU"
    "V7KQAtB+jkW4ktqWY+SUAMSC6RekcaZYhtATcWKP9n/T2uXfjYh2TYsinslGCsDh0QWxLqAF"
    "S3nCrzVR8CG2C9dj07PhMphCRFRfZJp6a5d/BeLwDvsdPWUSUgAOnxzEDsGg8M8KYqEp0hvY"
    "C3yCDCM2AyfC1R+OLsRa4U6trlszwuXfYLp1NkMKQPL4AWJKoFVDciK8gchCJusRuwXGBQYl"
    "h0MPhLuvJa+piK28BvRRfycwG72Sb1YjBSC5FCGyFAdFvJeL2DXQdgqagbWIuHIpBMmhDJHB"
    "1yPivSZEJw+Ff24ElpIGx3WZiRSA1NAHGIceXqogpgSRIdV+4BtgDXKhsL10R7j6XSLeCyBS"
    "0bWtWBURsv0h8ns+BCkAqcOJKCR5InrwkAshBJG5CUHEXHQ1YgtREp/eiI7fOeK9EOL7i3T3"
    "dyFyWTI2lv9wkQKQeooRySZHRLznRAhBXsR7IUShia+R89NoKIhKUMNpmaIeRMzzm9A7fhNi"
    "xF+F3H2JiRQA8+iHSKrqGvGeEzEtyKNlwdMqhBhsAcOjCLOFjgjx7EdLwQygd/zI91Yhdlws"
    "S09PJ6QAmE9vxLQgcn/agRCCfFoKQRBxXsK3iNXrbPnPykN0+CM4tNCLH9HxI8OtfcBKROCV"
    "7PhtQAqAdXRDCEH/iPcUxPpAHoeeg9CI8Ao2kZkpqk6EKB6BWNyLFEItN7+JljsnjYht1ZXI"
    "/It2IQXAekoRQjCIlje9EyEEeRxa4ageMU2oQngGDaQfCuLf3hUhhmUc+u9sRnTyZlp6PweB"
    "zxHJO1mTuJMKpADYhw6IvezBHHr2gQfdM4h23MkBdDGowp7bXQrCne+G6PRdiH44rR9hfyP6"
    "Hj4IAdiG2Db9hiwO300mUgDsh4JYJxiKqE3vanXNE37kYHy6cy1imlCL2FHQns0aLfMRK/XF"
    "Ec+lRD/eTUW49T7ESN/axn2ITr+WND+Fx45IAbA3boQIDEXMj1uP/g50QfAQ/7j3BoQY1CLc"
    "aB9ixNXOt498BNBHWXf44Yp4HfleHi07eyw7tEKbzRGf35pG9CApuYefQqQApA9FwJGIKMMe"
    "RO9kkYLgQsypD6dCcoj2nf2ooSKERBMTrcNHu+lqEC7+JkSR1lCUNpIkIwUgPXEi5tG9ENOF"
    "bhh3dCV8zYUuCtrz4R2lLFARnVU7NivyEasTH0B0+G2IrU7p3luAFIDMwIXYOuuNEIVS9KzE"
    "eCitHo5WP4Po5FpHj3xuy81zALFIuRXR6TNxKzPtkAKQuRQhVt07Is4z6BR+XRzrlw4T7TSd"
    "/YjFu8jnbI9otCVSALIPF0IIihBegrZmEO21GzHaa3N3n8HrJsQcXrrxaYYUAIkki0nGIpBE"
    "IklTpABIJFmMFACJJIuRAiCRZDFSACSSLEYKgESSxUgBkEiyGCkAEkkWIwVAIslipABIJFmM"
    "FACJJIuRAiCRZDFSACSSLEYKgESSxUgBkEiyGCkAEkkWIwVAIslipABIJFmMFACJJIuRAiCR"
    "ZDFSACSSLEYKgESSxUgBkEiyGCkAEkkWIwVAIslipABIJFmMFACJJIuRAiCRZDH/D8PLUIX+"
    "BVXhAAAAAElFTkSuQmCC")
getMainData = Main.GetData
getMainImage = Main.GetImage
getMainBitmap = Main.GetBitmap
getMainIcon = Main.GetIcon




# -*- coding: utf-8 -*-
""" PyCorrFit data set

Classes for FCS data evaluation.
"""
from __future__ import print_function, division

import copy
import hashlib
import lmfit
import numpy as np
import scipy.integrate as spintg
import scipy.interpolate as spintp
import warnings

from . import models as mdls

class Trace(object):
    """ unifies trace handling
    """
    def __init__(self, trace=None, countrate=None, duration=None,
                 name=None):
        """ Load trace data
        
        Parameters
        ----------
        trace : ndarray of shape (N, 2)
            The array contains time [ms] and count rate [kHz].
        coutrate : float
            Average count rate [kHz].
            Mandatory if `trace` is None. 
        duration : float
            Duration of measurement in milliseconds.
            Mandatory if `trace` is None.
        name : str
            The name of the trace.
        """
        self._countrate = None
        self._duration = None
        self._trace = None
        self._uid = None
        
        if trace is None:
            self.countrate = countrate
            self.duration = duration
        else:
            self.trace = trace
        
        if name is None:
            name = "{:.2f}kHz, {:.0f}s".format(self.countrate,
                                               self.duration/1000)
        self.name = name
    
    def __getitem__(self, idx):
        return self.trace[idx]
    
    def __repr__(self):
        text = "Trace of length {:.3f}s and countrate {:.3f}kHz".format(
                self.duration/1000, self.countrate)
        return text
    
    @property
    def countrate(self):
        if self._countrate is None:
            #self._countrate = np.average(self._trace[:,1])
            # Take into account traces that have arbitrary sampling
            self._countrate = spintg.simps(self._trace[:,1], self._trace[:,0]) / self.duration
        return self._countrate
    
    @countrate.setter
    def countrate(self, value):
        assert value is not None, "Setting value with None forbidden!"
        assert self._trace is None, "Setting value impossible, "+\
                                    "if `self.trace` is set."
        self._countrate = value

    @property
    def duration(self):
        if not hasattr(self, "_duration") or self._duration is None:
            self._duration = self._trace[-1,0] - self._trace[0,0]
        return self._duration
    
    @duration.setter
    def duration(self, value):
        assert value is not None, "Setting value with None forbidden!"
        assert self._trace is None, "Setting value impossible, "+\
                                    "if `self.trace` is set."
        self._duration = value
    
    @property
    def uid(self):
        if self._uid is None:
            hasher = hashlib.sha256()
            hasher.update(str(np.random.random()))
            hasher.update(str(self.trace))
            hasher.update(self.name)
            self._uid = hasher.hexdigest()
        return self._uid
    
    @property
    def trace(self):
        if self._trace is None:
            self._trace = np.array([ [0,             self.countrate],
                                     [self.duration, self.countrate] 
                                    ])
        return self._trace
    
    @trace.setter
    def trace(self, value):
        assert value is not None, "Setting value with None forbidden!"
        assert isinstance(value, np.ndarray), "value must be array!"
        assert value.shape[1] == 2, "shape of array must be (N,2)!"
        self._trace = value
        # self.countrate is set automagically


class Correlation(object):
    """ unifies correlation curve handling
    """
    def __init__(self, backgrounds=[], correlation=None, corr_type="AC", 
                 filename=None, fit_algorithm="Lev-Mar",
                 fit_model=6000, fit_ival=(0,0),
                 fit_weight_data=None, fit_weight_type="none", 
                 normparm=None, title=None, traces=[], verbose=1):
        """
        Parameters
        ----------
        backgrounds: list of instances of Trace
            background traces
        correlation: ndarray of shape (N,2)
            correlation data (time [s], correlation)
        corr_type: str
            type of correlation, e.g. "AC", "AC1", "cc12"
        filename: str
            path to filename of correlation
        fit_algorithm: str
            valid fit algorithm identifier (Algorithms.keys())
        fit_ival:
            fitting interval of lag times in indices
        fit_model: instance of FitModel
            the model used for fitting
        fit_weight_data: any
            data for the certain fit_weight_type
        fit_weight_type: str
            Reserved keywords or user-defined strings:
             - "none" : no weights are used
             - "splineX" : compute weights from spline with X knots
                   and a spread of `fit_weight_data` bins.
             - "model function" : compute weights from difference
                   to model function
             - user-defined : other weights (e.g. previously computed 
                   averages given in fit_weight_data)
        normparm: int
            identifier of normalization parameter
        title: str
            user-editable title of this correlation
        traces: list of instances of Trace
            traces of the current correlation
        verbose : int
            increment to increase verbosity
        """
        # must be created before setting properties
        self._backgrounds = []
        self._correlation = None
        self._fit_algorithm = None   
        self._fit_model = None
        self._fit_parameters = None
        self._fit_parameters_range = None
        self._fit_parameters_variable = None
        self._fit_weight_memory = dict()
        self._lag_time = None
        self._model_memory = dict()
        self._traces = []
        self._uid = None

        self.verbose = verbose

        self.backgrounds = backgrounds
        self.bg_correction_enabled = True
        self.correlation = correlation
        self.corr_type = corr_type
        self.filename = filename
        
        self.fit_algorithm = fit_algorithm
        self.fit_ival = fit_ival
        self.fit_model = fit_model
        # Do not change order:
        self.fit_weight_type = fit_weight_type
        self.fit_weight_parameters = fit_weight_data
    
        # lock prevents any changes to the parameters
        self.lock_parameters = False
        self.normparm = normparm
        self.title = title
        self.traces = traces

    def __repr__(self):
        if self.is_ac:
            c = "AC"
        else:
            c = "CC"
        text = "{} correlation '{}' with {} traces".format(
                c, self.title, len(self._traces))
        return text


    def background_replace(self, channel, background):
        """
        Replace a background.
        Channel must be 0 or 1.
        background must be instance of `Trace`
        """
        assert channel in [0, 1]
        assert isinstance(background, Trace)
        
        if self.is_ac:
            if channel == 1:
                raise ValueError("Cannot set second background for AC.")
            self._backgrounds = [background]
        else:
            if len(self._backgrounds) == 0:
                self._backgrounds = [Trace(countrate=0, duration=0), Trace(countrate=0, duration=0)]
            elif len(self._backgrounds) == 1:
                self._backgrounds.append(Trace(countrate=0, duration=0))
            self._backgrounds[channel] = background

    @property
    def backgrounds(self):
        """
        The background trace(s) of this correlation in a list.
        """
        return self._backgrounds
    
    @backgrounds.setter
    def backgrounds(self, value):
        """
        Set the backgrounds. The value can either be a list of traces or
        instances of traces or a single trace in an array.
        """
        backgrounds = []
        if not isinstance(value, list):
            value = [value]
        assert len(value) in [0,1,2], "Backgrounds must be list with up to two elements."
        for v in value:
            if isinstance(v, np.ndarray):
                backgrounds.append(Trace(trace=v))
            elif isinstance(v, Trace):
                backgrounds.append(v)
            else:
                raise ValueError("Each background must be instance of Trace or ndarray")
        self._backgrounds = backgrounds


    @property
    def bg_correction_factor(self):
        """
        Returns background correction factor for
        self._correlation
        
        Notes
        -----
        Thompson, N. Lakowicz, J.;
        Geddes, C. D. & Lakowicz, J. R. (ed.)
        Fluorescence Correlation Spectroscopy
        Topics in Fluorescence Spectroscopy,
        Springer US, 2002, 1, 337-378
        """
        if not self.bg_correction_enabled:
            # bg correction disabled
            return 1

        if self.is_ac:
            # Autocorrelation
            if len(self.traces) == 1 and len(self.backgrounds) == 1:
                S = self.traces[0].countrate
                B = self.backgrounds[0].countrate
                bgfactor = (S/(S-B))**2
            else:
                if self.verbose >= 1:
                    warnings.warn("Correlation {}: no bg-correction".
                                  format(self.uid))
                bgfactor = 1
        else:
            # Crosscorrelation
            if len(self.traces) == 2 and len(self.backgrounds) == 2:
                S = self.traces[0].countrate
                S2 = self.traces[1].countrate
                B = self.backgrounds[0].countrate
                B2 = self.backgrounds[1].countrate
                bgfactor = (S/(S-B)) * (S2/(S2-B2))
            else:
                warnings.warn("Correlation {}: no bg-correction".
                              format(self))
                bgfactor = 1
        return bgfactor

    def check_parms(self, parms):
        """ Check parameters using self.fit_model.func_verification and the user defined
            boundaries self.fit_parameters_range for each parameter.
        """
        p = 1.*np.array(parms)
        r = self.fit_parameters_range
        for i in range(len(p)):
            if r[i][0] == r[i][1]:
                pass
            elif r[i][0] is None:
                if p[i] > r[i][1]:
                    p[i] = r[i][1]
            elif r[i][1] is None:
                if p[i] < r[i][0]:
                    p[i] = r[i][1]
            elif p[i] < r[i][0]:
                p[i] = r[i][0]
            elif p[i] > r[i][1]:
                p[i] = r[i][1]
        return p

    @property
    def correlation(self):
        """the correlation data, shape (N,2) with (time, correlation) """
        if self._correlation is not None:
            corr = self._correlation.copy()
            return corr
    
    @correlation.setter
    def correlation(self, value):
        if value is None:
            warnings.warn("Setting correlation to `None`.")
        elif not isinstance(value, np.ndarray):
            raise ValueError("Correlation must be 2d array!")
        elif not value.shape[1] == 2:
            raise ValueError("Correlation array must have shape (N,2)!")
        self._correlation = value

    @property
    def correlation_fit(self):
        """ returns correlation data for fitting (fit_ivald)
        - background correction
        - fitting interval cropping
        """
        corr = self.correlation
        if corr is not None:
            # perform background correction
            corr[:,1] *= self.bg_correction_factor
            # perform parameter normalization
            return corr[self.fit_ival[0]:self.fit_ival[1],:]
        
    
    @property
    def correlation_plot(self):
        """ returns correlation data for plotting (normalized, fit_ivald)
        - background correction
        - fitting interval cropping
        - parameter normalization
        """
        corr = self.correlation_fit
        if corr is not None:
            # perform parameter normalization
            corr[:,1] *= self.normalize_factor
            return corr

    @property
    def is_ac(self):
        """True if instance contains autocorrelation"""
        return self.corr_type.lower().count("ac") > 0

    @property
    def is_cc(self):
        """True if instance contains crosscorrelation"""
        return not self.is_ac

    @property
    def is_weighted_fit(self):
        """True if a weighted fit was performed"""
        return self.fit_weight_type != "none"

    @property
    def fit_algorithm(self):
        """The string representing the fitting algorithm"""
        return self._fit_algorithm

    @fit_algorithm.setter
    def fit_algorithm(self, value):
        # TODO:
        # - allow lower-case fitting algorithm
        assert value in list(Algorithms.keys()), "Invalid fit algorithm: "+value
        self._fit_algorithm = value

    @property
    def fit_model(self):
        """instance of a fit model"""
        return self._fit_model

    @fit_model.setter
    def fit_model(self, value):
        """set the fit model
        """
        if isinstance(value, (int, long)):
            newmodel = mdls.modeldict[value]
        elif isinstance(value, mdls.Model):
            newmodel = value
        else:
            raise NotImplementedError("Unknown model identifier")
        
        if newmodel != self._fit_model :
            self._fit_model = newmodel
            # overwrite fitting parameters
            self._fit_parameters = self._fit_model.default_values
            self._fit_parameters_variables = self._fit_model.default_variables
            self._fit_parameters_range = np.zeros((len(self._fit_parameters), 2))
            self.normalize_parm = None

    @property
    def fit_ival(self):
        """lag time interval for fitting"""
        lag = self.lag_time
        if lag is not None:
            if self._fit_ival[1] <= 0 or self._fit_ival[1] > lag.shape[0]:
                self._fit_ival[1] = lag.shape[0]
        return self._fit_ival
    
    @fit_ival.setter
    def fit_ival(self, value):
        value = list(value)
        if value[1] <= 0:
            if self.lag_time is not None:
                value[1] = self.lag_time.shape[0]
            else:
                # just to be sure
                warnings.warn("No data available.")
                value[1] = 10000000000000000
        self._fit_ival = value

    @property
    def fit_weight_data(self):
        """data of weighted fitting"""
        try:
            data = self._fit_weight_memory[self.fit_weight_type]
        except KeyError:
            # Standard variables for weights
            if self.fit_weight_type.count("spline"):
                # Default area for weighting with spline fit
                data = 3
            else:
                data = None
        return data

    @fit_weight_data.setter
    def fit_weight_data(self, value):
        self._fit_weight_memory[self.fit_weight_type] = value

    @property
    def fit_parameters(self):
        """parameters that were fitted/will be used for fitting"""
        return self._fit_parameters

    @fit_parameters.setter
    def fit_parameters(self, value):
        # must unlock parameters, if change is required
        value = np.array(value)
        if self.lock_parameters == False:
            self._fit_parameters = self.check_parms(value)
        else:
            warnings.warn("Correlation {}: fixed parameters unchanged.".
                          format(self.uid))

    @property
    def fit_parameters_range(self):
        """valid fitting ranges for fit parameters"""
        model = self.fit_model.boundaries
        mine = self._fit_parameters_range
        new = []
        for a, b in zip(model, mine):
            c = [-np.inf, np.inf]
            if a[0] != a[1]:
                c[0] = a[0]
                c[1] = a[1]
            # user overrides model
            if b[0] != b[1]:
                c[0] = b[0]
                c[1] = b[1]
            if c[0] is not None and np.isnan(c[0]):
                c[0] = -np.inf
            if c[1] is not None and np.isnan(c[1]):
                c[1] = np.inf

            new.append(c)         
        return np.array(new)

    @fit_parameters_range.setter
    def fit_parameters_range(self, value):
        value = np.array(value)
        assert value.shape[1] == 2
        assert value.shape[0] == self.fit_parameters.shape[0]
        self._fit_parameters_range = value

    @property
    def fit_parameters_variable(self):
        """which parameters are variable during fitting"""
        if self._fit_parameters_variable is None:
            self._fit_parameters_variable = np.array(self.fit_model.default_variables, dtype=bool)
        return self._fit_parameters_variable

    @fit_parameters_variable.setter
    def fit_parameters_variable(self, value):
        value = np.array(value, dtype=bool)
        assert value.shape[0] == self.fit_parameters.shape[0]
        self._fit_parameters_variable = value

    @property
    def lag_time(self):
        """logarithmic lag time axis"""
        if self.correlation is not None:
            return self._correlation[:,0].copy()
        elif self._lag_time is not None:
            return self._lag_time
        else:
            # some default lag time
            return 10**np.linspace(-6,8,1001)

    @lag_time.setter
    def lag_time(self, value):
        if self.correlation is not None:
            warnings.warn("Setting lag time not possible, because of existing correlation")
        else:
            self._lag_time = value

    @property
    def lag_time_fit(self):
        """lag time as used for fitting"""
        return self.lag_time[self.fit_ival[0]:self.fit_ival[1]]

    @property
    def modeled(self):
        """fitted data values, same shape as self.correlation"""
        # perform parameter normalization
        lag = self.lag_time
        modeled = np.zeros((lag.shape[0], 2))
        modeled[:,0] = lag
        modeled[:,1] = self.fit_model(self.fit_parameters, lag)
        return modeled.copy()

    @property
    def modeled_fit(self):
        """fitted data values, same shape as self.correlation_fit"""
        toplot = self.modeled[self.fit_ival[0]:self.fit_ival[1], :]
        return toplot

    @property
    def modeled_plot(self):
        """fitted data values, same shape as self.correlation_fit"""
        toplot = self.modeled_fit
        toplot[:,1] *= self.normalize_factor
        if toplot[-1,1] == 0.1:
            import IPython
            IPython.embed()
        return toplot

    @property
    def normalize_factor(self):
        """plot normalization according to self.normparm"""
        if self.normparm is None:
            # nothing to do
            return 1
        
        if self.normparm < self.fit_parameters.shape[0]:
            nfactor = self.fit_parameters[self.normparm]
        else:
            # get supplementary parameters
            alt = self.fit_model.get_supplementary_values(self.fit_parameters)
            nfactor = alt[self.normparm - self.fit_parameters.shape[0]]
        
        return nfactor

    @property
    def residuals(self):
        """fit residuals, same shape as self.correlation"""
        if self.correlation is None:
            raise ValueError("Cannot compute residuals; No correlation given!") 
        residuals = self.correlation.copy()
        residuals[:,1] -= self.modeled[:,1]
        return residuals 
    
    @property
    def residuals_fit(self):
        """fit residuals, same shape as self.correlation_fit"""
        residuals_fit = self.correlation_fit.copy()
        residuals_fit[:,1] -= self.modeled_fit[:,1]
        return residuals_fit

    @property
    def residuals_plot(self):
        """fit residuals, same shape as self.correlation_fit"""
        cp = self.correlation_plot
        if cp is not None:
            residuals_plot = self.correlation_plot.copy()
            residuals_plot[:,1] -= self.modeled_plot[:,1]
            return residuals_plot

    def set_weights(self, type_name, data):
        """
        Add weights for fitting.
        example:
        type_name : "Average"
        data : 1d ndarray with length self.lag_time 
        """
        if data is not None:
            self._fit_weight_memory[type_name] = data

    @property
    def traces(self):
        """
        The trace(s) of this correlation in a list.
        """
        return self._traces
    
    @traces.setter
    def traces(self, value):
        """
        Set the traces. The value can either be a list of traces or
        instances of traces or a single trace in an array.
        """
        traces = []
        if not isinstance(value, list):
            value = [value]
        assert len(value) in [0,1,2], "Traces must be list with up to two elements."
        for v in value:
            if isinstance(v, np.ndarray):
                traces.append(Trace(trace=v))
            elif isinstance(v, Trace):
                traces.append(v)
            else:
                raise ValueError("Each trace must be instance of Trace or ndarray")
        self._traces = traces
        
        if len(self._traces) == 2:
            if self._traces[0].duration != self._traces[1].duration:
                warnings.warn("Unequal lenght of traces: {} and {}".format(
                              self._traces[0].duration,
                              self._traces[1].duration))

    @property
    def uid(self):
        """
        unique identifier of this instance
        This might change when title or filename
        are updated.
        """
        if self._uid is None:
            hasher = hashlib.sha256()
            hasher.update(str(np.random.random()))
            hasher.update(str(self._correlation))
            hasher.update(str(self.filename))
            hasher.update(str(self.title))
            self._uid = hasher.hexdigest()
        return self._uid


class Fit(object):
    """ Used for fitting FCS data to models.
    """
    def __init__(self, correlations=[], global_fit=False,
                 global_fit_variables=[],
                 uselatex=False, verbose=0):
        """ Using an FCS model, fit the data of shape (N,2).


        Parameters
        ----------
        correlations: list of instances of Correlation
            Correlations to fit.
        global fit : bool
            Perform global fit. The default behavior is
            to fit all parameters that are selected for
            fitting in each correlation. Parameters with
            the same name in different models are treated
            as one global parameter. 
        global_fit_variables: list of list of strings
            Each item contains a list of strings that are names
            of parameters which will be treated as a common
            parameter. This breaks the defaul behavior.
            NOT IMPLEMENTED YET!
        verbose: int
            Increase verbosity by incrementing this number.
        uselatex: bool
            If verbose > 0, plotting will be performed with LaTeX.
        """
        assert len(global_fit_variables)==0, "not implemented"
        
        if isinstance(correlations, Correlation):
            correlations = [correlations]
        
        self.correlations = correlations
        self.global_fit_variables = global_fit_variables
        self.verbose = verbose
        self.uselatex = uselatex
        self.is_weighted_fit = False
        
        if not global_fit:
            # Fit each correlation separately
            for corr in self.correlations:
                # Set fitting options
                self.fit_algorithm = corr.fit_algorithm
                # Get the data required for fitting
                self.x = corr.correlation_fit[:,0]
                self.y = corr.correlation_fit[:,1]
                # fit_bool: True for variable
                self.fit_bool = corr.fit_parameters_variable.copy()
                self.fit_parm = corr.fit_parameters.copy()
                self.fit_bound = copy.copy(corr.fit_parameters_range)
                self.is_weighted_fit = corr.is_weighted_fit
                self.fit_weights = Fit.compute_weights(corr,
                                                   verbose=verbose,
                                                   uselatex=uselatex)
                self.fit_parm_names = corr.fit_model.parameters[0]
                self.func = corr.fit_model.function
                self.check_parms = corr.check_parms
                self.constraints = corr.fit_model.constraints
                # Directly perform the fit and set the "fit" attribute
                self.minimize()
                # Run a second time:
                self.minimize()
                # update correlation model parameters
                corr.fit_parameters = self.fit_parm
                # save fit instance in correlation class
                corr.fit_results = self.get_fit_results(corr)
        else:
            # TODO:
            # - allow detaching of parameters,
            #   i.e. fitting "n" separately for two models
            # Initiate all arrays
            self.fit_algorithm = self.correlations[0].fit_algorithm
            xtemp = []      # x
            ytemp = []      # y
            weights = []    # weights
            ids = [0]           # ids in big fitting array
            cmodels = []    # correlation model info
            initpar = []    # initial parameters
            varin = []      # names of variable fitting parameters
            variv = []      # values of variable fitting parameters
            varmap = []     # list of indices of fitted parameters
            varbound = []   # list of fitting boundaries
            self.is_weighted_fit = None
            for corr in self.correlations:
                xtemp.append(corr.correlation_fit[:,0])
                ytemp.append(corr.correlation_fit[:,1])
                weights.append(Fit.compute_weights(corr))
                ids.append(len(xtemp[-1])+ids[-1])
                cmodels.append(corr.fit_model)
                initpar.append(corr.fit_parameters)
                # Create list of variable parameters
                varthis = []
                for ipm, par in enumerate(corr.fit_model.parameters[0]):
                    if corr.fit_parameters_variable[ipm]:
                        varthis.append(ipm)
                        varin.append(par)
                        variv.append(corr.fit_parameters[ipm])
                        varbound.append(corr.fit_parameters_range[ipm])
                varmap.append(varthis)

            # These are the variable fitting parameters
            __, varidx = np.unique(varin, return_index=True)
            varidx.sort()
            varin = np.array(varin)[varidx]
            variv = np.array(variv)[varidx]
            
            self.x = np.concatenate(xtemp)
            self.y = np.concatenate(ytemp)
            self.fit_bool = np.ones(len(variv), dtype=bool)
            self.fit_parm = variv
            self.fit_weights = np.concatenate(weights)
            self.fit_parm_names = varin
            self.fit_bound = varbound
            self.constraints = []
            warnings.warn("Constraints are not supported yet for global fitting.")
            
            
            def parameters_global_to_local(parameters, iicorr, varin=varin,
                                          initpar=initpar,
                                          correlations=correlations):
                """
                With global `parameters` and an id `iicorr` pointing at
                the correlation in `self.correlations`, return the
                updated parameters of the corresponding model.
                """
                fit_parm = initpar[iicorr].copy()
                corr = correlations[iicorr]
                mod = corr.fit_model
                for kk, pn in enumerate(mod.parameters[0]):
                    if pn in varin:
                        # edit that parameter
                        fit_parm[kk] = parameters[np.where(np.array(varin)==pn)[0]]
                return fit_parm
            
            def parameters_local_to_global(parameters, iicorr, fit_parm,
                                           varin=varin,
                                           correlations=correlations):
                """
                inverse of parameters_global_to_local
                """
                corr = correlations[iicorr]
                mod = corr.fit_model
                for kk, pn in enumerate(mod.parameters[0]):
                    if pn in varin:
                        # edit that parameter
                        parameters[np.where(np.array(varin)==pn)[0]] = fit_parm[kk]
                return parameters
            
            # Create function for fitting using ids
            def global_func(parameters, tau,
                            glob2loc=parameters_global_to_local):
                out = []
                # ids start at 0
                for ii, mod in enumerate(cmodels):
                    # Update parameters
                    fit_parm = glob2loc(parameters, ii)
                    # return function
                    out.append(mod.function(fit_parm, tau[ids[ii]:ids[ii+1]]))
                return np.concatenate(out)

            self.func = global_func
            
            # Create function for checking
            def global_check_parms(parameters,
                                   glob2loc=parameters_global_to_local,
                                   loc2glob=parameters_local_to_global):

                for ii, corr in enumerate(self.correlations):
                    # create new initpar
                    fit_parm = glob2loc(parameters, ii)
                    fit_parm = corr.check_parms(fit_parm)
                    # update parameters
                    parameters = loc2glob(parameters, ii, fit_parm)

                return parameters
            
            self.check_parms = global_check_parms

            # Directly perform the fit and set the "fit" attribute
            self.minimize()
            # Update correlations
            for ii, corr in enumerate(self.correlations):
                # write new model parameters
                corr.fit_parameters = parameters_global_to_local(self.fit_parm,
                                                                 ii)
                # save fit instance in correlation class
                corr.fit_results = self.get_fit_results(corr)


    def get_fit_results(self, correlation):
        """
        Return a dictionary with all information about the performed fit.
        
        This function must be called immediately after `self.minimize`.
        """
        c = correlation
        d = {
             "chi2" : self.chi_squared,
             "chi2 type" : self.chi_squared_type,
             "weighted fit" : c.is_weighted_fit,
             "fit algorithm" : c.fit_algorithm,
             "fit result" : c.fit_parameters.copy(),
             "fit parameters" : np.where(c.fit_parameters_variable)[0],
             "fit weights" : self.compute_weights(c)
             }
        
        
        if c.is_weighted_fit:
            d["weighted fit type"] = c.fit_weight_type
            if isinstance(c.fit_weight_data, (int, float)):
                d["weighted fit bins"] = c.fit_weight_data

        if d["fit algorithm"] == "Lev-Mar" and self.parmoptim_error is not None:
            d["fit error estimation"] = self.parmoptim_error
        
        
        return d
        

    @property
    def chi_squared(self):
        """ Calculate displayed Chi²
        
            Calculate reduced Chi² for the current class.
        """
        # Calculate degrees of freedom
        dof = len(self.x) - np.sum(self.fit_bool) - 1
        # This is exactly what is minimized by the scalar minimizers
        
        if self.chi_squared_type == "reduced expected sum of squares":
            fitted = self.func(self.fit_parm, self.x)
            chi2 = np.sum((self.y-fitted)**2/np.abs(fitted)) / dof
        elif self.chi_squared_type == "reduced weighted sum of squares":
            fitted = self.func(self.fit_parm, self.x)
            variance = self.fit_weights**2
            chi2 = np.sum((self.y-fitted)**2/variance) / dof
        elif self.chi_squared_type == "reduced global sum of squares":
            fitted = self.func(self.fit_parm, self.x)
            variance = self.fit_weights**2
            chi2 = np.sum((self.y-fitted)**2/variance) / dof
        else:
            chi2 = self.fit_function_scalar(self.fit_parm, self.x, self.y, self.fit_weights)
        
            
        return chi2


    @property
    def chi_squared_type(self):
        """ The type of Chi² that currently applies.
        
        Returns
        -------
        "reduced" - if variance of data was used for fitting
        "reduced Pearson" - if variance of data is not available
        """
        if self.is_weighted_fit is None:
            # global fitting
            return "reduced global sum of squares"
        elif self.is_weighted_fit == True:
            return "reduced weighted sum of squares"
        elif self.is_weighted_fit == False:
            return "reduced expected sum of squares"
        else:
            raise ValueError("Unknown weight type!")


    @staticmethod
    def compute_weights(correlation, verbose=0, uselatex=False):
        """ computes and returns weights of the same length as 
        `correlation.correlation_fit`
        
        `correlation` is an instance of Correlation
        """
        corr = correlation
        model = corr.fit_model
        model_parms = corr.fit_parameters
        ival = corr.fit_ival
        weight_data = corr.fit_weight_data
        weight_type = corr.fit_weight_type
        #parameters = corr.fit_parameters
        #parameters_range = corr.fit_parameters_range
        #parameters_variable = corr.fit_parameters_variable
        
        cdat = corr.correlation
        if cdat is None:
            raise ValueError("Cannot compute weights; No correlation given!")
        cdatfit = corr.correlation_fit
        x_full = cdat[:,0]
        y_full = cdat[:,1]
        x_fit = cdatfit[:,0]
        #y_fit = cdatfit[:,1]
        
        dataweights = np.ones_like(x_fit)

        try:
            weight_spread = int(weight_data)
        except:
            if verbose > 1:
                warnings.warn("Could not get weight spread for spline. Setting it to 3.")
            weight_spread = 3

        if weight_type[:6] == "spline":
            # Number of knots to use for spline
            try:
                knotnumber = int(weight_type[6:])
            except:
                if verbose > 1:
                    print("Could not get knot number. Setting it to 5.")
                knotnumber = 5
            
            # Compute borders for spline fit.
            if ival[0] < weight_spread:
                # optimal case
                pmin = ival[0]
            else:
                # non-optimal case
                # we need to cut pmin
                pmin = weight_spread
            if x_full.shape[0] - ival[1] < weight_spread:
                # optimal case
                pmax = x_full.shape[0] - ival[1]
            else:
                # non-optimal case
                # we need to cut pmax
                pmax = weight_spread

            x = x_full[ival[0]-pmin:ival[1]+pmax]
            y = y_full[ival[0]-pmin:ival[1]+pmax]
            # we are fitting knots on a base 10 logarithmic scale.
            xs = np.log10(x)
            knots = np.linspace(xs[1], xs[-1], knotnumber+2)[1:-1]
            try:
                tck = spintp.splrep(xs, y, s=0, k=3, t=knots, task=-1)
                ys = spintp.splev(xs, tck, der=0)
            except:
                if verbose > 0:
                    raise ValueError("Could not find spline fit with "+\
                                     "{} knots.".format(knotnumber))
                return
            if verbose > 0:
                ## If plotting module is available:
                #name = "spline fit: "+str(knotnumber)+" knots"
                #plotting.savePlotSingle(name, 1*x, 1*y, 1*ys,
                #                         dirname=".",
                #                         uselatex=uselatex)
                # use matplotlib.pylab
                try:
                    from matplotlib import pylab as plt
                    plt.xscale("log")
                    plt.plot(x, ys, x, y)
                    plt.show()
                except ImportError:
                    # Tell the user to install matplotlib
                    print("Couldn't import pylab! - not Plotting")

            ## Calculation of variance
            # In some cases, the actual cropping interval from ival[0]
            # to ival[1] is chosen, such that the dataweights must be
            # calculated from unknown datapoints.
            # (e.g. points+endcrop > len(correlation)
            # We deal with this by multiplying dataweights with a factor
            # corresponding to the missed points.
            for i in range(x_fit.shape[0]):
                # Define start and end positions of the sections from
                # where we wish to calculate the dataweights.
                # Offset at beginning:
                if  i + ival[0] <  weight_spread:
                    # The offset that occurs
                    offsetstart = weight_spread - i - ival[0]
                    offsetcrop = 0
                elif ival[0] > weight_spread:
                    offsetstart = 0
                    offsetcrop = ival[0] - weight_spread
                else:
                    offsetstart = 0
                    offsetcrop = 0
                # i: counter on correlation array
                # start: counter on y array
                start = i - weight_spread + offsetstart + ival[0] - offsetcrop
                end = start + 2*weight_spread + 1 - offsetstart
                dataweights[i] = (y[start:end] - ys[start:end]).std()
                # The standard deviation at the end and the start of the
                # array are multiplied by a factor corresponding to the
                # number of bins that were not used for calculation of the
                # standard deviation.
                if offsetstart != 0:
                    reference = 2*weight_spread + 1
                    dividor = reference - offsetstart
                    dataweights[i] *= reference/dividor   
                # Do not substitute len(y[start:end]) with end-start!
                # It is not the same!
                backset =  2*weight_spread + 1 - len(y[start:end]) - offsetstart
                if backset != 0:
                    reference = 2*weight_spread + 1
                    dividor = reference - backset
                    dataweights[i] *= reference/dividor
        elif weight_type == "model function":
            # Number of neighboring (left and right) points to include
            if ival[0] < weight_spread:
                pmin = ival[0]
            else:
                pmin = weight_spread
            if x_full.shape[0] - ival[1] <  weight_spread:
                pmax = x_full.shape[0] - ival[1]
            else:
                pmax = weight_spread
            x = x_full[ival[0]-pmin:ival[1]+pmax]
            y = y_full[ival[0]-pmin:ival[1]+pmax]
            # Calculated dataweights
            for i in np.arange(x_fit.shape[0]):
                # Define start and end positions of the sections from
                # where we wish to calculate the dataweights.
                # Offset at beginning:
                if  i + ival[0] <  weight_spread:
                    # The offset that occurs
                    offsetstart = weight_spread - i - ival[0]
                    offsetcrop = 0
                elif ival[0] > weight_spread:
                    offsetstart = 0
                    offsetcrop = ival[0] - weight_spread
                else:
                    offsetstart = 0
                    offsetcrop = 0
                # i: counter on correlation array
                # start: counter on correlation array
                start = i - weight_spread + offsetstart + ival[0] - offsetcrop
                end = start + 2*weight_spread + 1 - offsetstart
                #start = ival[0] - weight_spread + i
                #end = ival[0] + weight_spread + i + 1
                diff = y - model(model_parms, x)
                dataweights[i] = diff[start:end].std()
                # The standard deviation at the end and the start of the
                # array are multiplied by a factor corresponding to the
                # number of bins that were not used for calculation of the
                # standard deviation.
                if offsetstart != 0:
                    reference = 2*weight_spread + 1
                    dividor = reference - offsetstart
                    dataweights[i] *= reference/dividor   
                # Do not substitute len(diff[start:end]) with end-start!
                # It is not the same!
                backset =  2*weight_spread + 1 - len(diff[start:end]) - offsetstart
                if backset != 0:
                    reference = 2*weight_spread + 1
                    dividor = reference - backset
                    dataweights[i] *= reference/dividor
        elif weight_type == "none":
            pass
        else:
            # This means that the user knows the dataweights and already
            # gave it to us.
            weights = weight_data
            assert weights is not None, "User defined weights not given: "+weight_type
            
            # Check if these other weights have length of the cropped
            # or the full array.
            if weights.shape[0] == x_fit.shape[0]:
                dataweights = weights
            elif weights.shape[0] == x_full.shape[0]:
                dataweights = weights[ival[0]:ival[1]]
            else:
                raise ValueError, \
                  "`weights` must have length of full or cropped array."
        
        return dataweights
        

    def fit_function(self, params, x, y, weights=1):
        """ 
        objective function that returns the residual (difference between
        model and data) to be minimized in a least squares sense.
        """
        parms = Fit.lmfitparm2array(params)
        tominimize = (self.func(parms, x) - y)
        # Check dataweights for zeros and don't use these
        # values for the least squares method.
        with np.errstate(divide='ignore'):
            tominimize = np.where(weights!=0, 
                                  tominimize/weights, 0)
        ## There might be NaN values because of zero weights:
        #tominimize = tominimize[~np.isinf(tominimize)]
        return tominimize

    def fit_function_scalar(self, parms, x, y, weights=1):
        """
        Wrapper of `fit_function`.
        Returns the sum of squares of the input data.
        """
        e = self.fit_function(parms, x, y, weights)
        return np.sum(e*e)

    def get_lmfitparm(self):
        """
        Generates an lmfit parameter class from the present data set.

        The following parameters are used:
        self.x : 1d ndarray length N
        self.y : 1d ndarray length N
        self.fit_weights : 1d ndarray length N
        
        self.fit_bool : 1d ndarray length P, bool
        self.fit_parm : 1d ndarray length P, float
        """
        params = lmfit.Parameters()

        # First, add all fixed parameters
        for pp in range(len(self.fit_parm)):
            if not self.fit_bool[pp]:
                if self.fit_bound[pp][0] == self.fit_bound[pp][1]:
                    self.fit_bound[pp]=[-np.inf, np.inf]
                params.add(lmfit.Parameter(name="parm{:04d}".format(pp),
                                           value=self.fit_parm[pp],
                                           vary=self.fit_bool[pp],
                                           min=self.fit_bound[pp][0],
                                           max=self.fit_bound[pp][1],
                                            )
                                           )
        
        # Second, summarize the constraints in a dictionary, where
        # keys are the parameter indexes of varied parameters.
        # The dictionary cstrnew only allows integer keys that are
        # representing parameter indices. The fact that we are effectively
        # reducing the number of valid constraints to one per parameter
        # is a design problem that cannot be resolved here. The constraints
        # must be defined in such a way, that a parameter with a larger
        # index number is dependent on only one parameter with a lower
        # index number, e.g. parm1>parm0, parm3<parm1, etc..
        cstrnew = {}
        bound = np.array(self.fit_bound).copy()
        for cc in self.constraints:
            if self.fit_bool[cc[0]] and self.fit_bool[cc[2]]:
                # Both cc[0] and c[2] are varied.
                # Everything will work fine, independent of the
                # the fact if cc[2] is varied or not.
                cstrnew[cc[0]] = [cc[1], cc[2]]
            elif self.fit_bool[cc[0]]:
                # Only cc[0] is varied, create boundary
                if cc[1] == "<":
                    # maximum
                    bnd = [-np.inf, self.fit_parm[cc[2]]]
                elif cc[1] == ">":
                    # minimum
                    bnd = [self.fit_parm[cc[2]], np.inf]
                # update boundaries if necessary
                bound[cc[0]] = [max(bound[cc[0]][0], bnd[0]),
                                min(bound[cc[0]][1], bnd[1])]
            elif self.fit_bool[cc[2]]:
                # Only cc[2] is varied, create boundary
                if cc[1] == "<":
                    # minimum boundary
                    bnd = [self.fit_parm[cc[0]], np.inf]
                elif cc[1] == ">":
                    # maximum boundary
                    bnd = [-np.inf, self.fit_parm[cc[0]]]
                # update boundaries if necessary
                bound[cc[2]] = [max(bound[cc[2]][0], bnd[0]),
                                min(bound[cc[2]][1], bnd[1])]
            else:
                # Neither cc[0] nor cc[2] are varied.
                # Do nothing.
                pass

        # Third, setup all variable parameters with the necessary constraints.
        for pp in range(len(self.fit_parm)):
            if self.fit_bool[pp]:
                # analyze constraints using lmfit:
                if pp in cstrnew:
                    # constrained parameters
                    ppref = cstrnew[pp][1]
                    rel = cstrnew[pp][0]
                    #TODO:
                    # - combine the two following cases for better readybility
                    if rel == "<":
                        #p2 < p1
                        #-> p2 = p1 - d21
                        #-> d21 = p1 - p2
                        #-> d21 > 0
                        deltaname="delta_{}_{}".format(pp, ppref)
                        params.add(lmfit.Parameter(name=deltaname,
                                               value=self.fit_parm[ppref]-self.fit_parm[pp],
                                               vary=self.fit_bool[pp],
                                               min=0,
                                               max=np.inf,
                                                ))
                        ppcomp = "parm{:04d}-{}".format(ppref, deltaname)
                        params.add(lmfit.Parameter(name="parm{:04d}".format(pp),
                                                         # this condition deals with negative numbers
                                                   expr="{MIN} if {COMP} < {MIN} else {MAX} if {COMP} > {MAX} else {COMP}".format(
                                                        COMP=ppcomp,
                                                        MIN=bound[pp][0],
                                                        MAX=bound[pp][1])
                                                ))
                    elif rel == ">":
                        # The opposite of the above case
                        #p2 > p1
                        #-> p2 = p1 + d21
                        #-> d21 = p2 - p1
                        #-> d21 > 0
                        deltaname="delta_{}_{}".format(pp, ppref)
                        params.add(lmfit.Parameter(name=deltaname,
                                               value=self.fit_parm[pp]-self.fit_parm[ppref],
                                               vary=self.fit_bool[pp],
                                               min=0,
                                               max=np.inf,
                                                ))
                        ppcomp = "parm{:04d}+{}".format(ppref, deltaname)
                        params.add(lmfit.Parameter(name="parm{:04d}".format(pp),
                                                         # this condition deals with negative numbers
                                                   expr="{MIN} if {COMP} < {MIN} else {MAX} if {COMP} > {MAX} else {COMP}".format(
                                                        COMP=ppcomp,
                                                        MIN=bound[pp][0],
                                                        MAX=bound[pp][1])
                                                ))
                    else:
                        raise NotImplementedError("Only '<' and '>' are allowed constraints!")
                
                else:
                    ## normal parameter
                    params.add(lmfit.Parameter(name="parm{:04d}".format(pp),
                                               value=self.fit_parm[pp],
                                               vary=self.fit_bool[pp],
                                               min=bound[pp][0],
                                               max=bound[pp][1],
                                                )
                                               )
        return params

    @staticmethod
    def lmfitparm2array(parms, parmid="parm", attribute="value"):
        """
        Convert lmfit parameters to a numpy array.
        Parameters are identified by name `parmid` which should
        be at the beginning of a parameters.
        
        This method is necessary to separate artificial constraint parameters
        from the actual parameters.
        
        Parameters
        ----------
        parms : lmfit.parameter.Parameters or ndarray
            The input parameters.
        parmid : str
            The identifier for parameters. By default this is
            "parm", i.e. parameters are named like this:
            "parm0001", "parm0002", etc.
        attribute : str
            The attribute to return, e.g.
            - "value" : return the current value of the parameter
            - "vary" : return if the parameter is varied during fitting
        
        Returns:
        parr : ndarray
            If the input is an ndarray, the input will be returned.
        """
        if isinstance(parms, lmfit.parameter.Parameters):
            items = parms.items()
            items.sort(key=lambda x: x[0])
            parr = [getattr(p[1], attribute) for p in items if p[0].startswith(parmid)]
        else:
            parr = parms

        return np.array(parr)

    def minimize(self):
        """ This will run the minimization process
      
        """
        assert (np.sum(self.fit_bool) != 0), "No parameter selected for fitting."
        
        # get all parameters for minimization
        params = self.get_lmfitparm()
        
        # Get algorithm
        method = Algorithms[self.fit_algorithm][0]
        methodkwargs = Algorithms[self.fit_algorithm][2]

        # Begin fitting
        # Fit a several times and stop earlier if the residuals
        # are small enough (heuristic approach).
        nfits = 5
        diff = np.inf
        parmsinit = Fit.lmfitparm2array(params)
        for ii in range(nfits):
            res0 = self.fit_function(params, self.x, self.y)
            result = lmfit.minimize(fcn=self.fit_function,
                                    params=params,
                                    method=method,
                                    kws={"x":self.x,
                                            "y":self.y,
                                            "weights":self.fit_weights},
                                    **methodkwargs
                                    )
            params = result.params
            res1 = self.fit_function(params, self.x, self.y)
            diff = np.average(np.abs(res0-res1))

            if hasattr(result, "ier") and not result.errorbars and ii+1 < nfits:
                # This case applies to the Levenberg-Marquardt algorithm
                multby = .5
                # Try to vary stuck fitting parameters
                # the result from the previous fit
                parmsres = Fit.lmfitparm2array(params)
                # the parameters that are varied during fitting
                parmsbool = Fit.lmfitparm2array(params, attribute="vary")
                # The parameters that are stuck
                parmstuck = parmsbool * (parmsinit==parmsres)
                parmsres[parmstuck] *= multby
                # write changes
                self.fit_parm = parmsres
                params = self.get_lmfitparm()
                warnings.warn(u"PyCorrFit detected problems in fitting, "+\
                              u"detected a stuck parameter, multiplied "+\
                              u"it by {}, and fitted again. ".format(multby)+\
                              u"The stuck parameters are: {}".format(
                                np.array(self.fit_parm_names)[parmstuck]))
            elif diff < 1e-8:
                # Experience tells us this is good enough.
                break

        # Now write the optimal parameters to our values:
        self.fit_parm = Fit.lmfitparm2array(params)
        # Only allow physically correct parameters
        self.fit_parm = self.check_parms(self.fit_parm)
        # Compute error estimates for fit (Only "Lev-Mar")
        if self.fit_algorithm == "Lev-Mar" and result.success:
            # This is the standard way to minimize the data. Therefore,
            # we are a little bit more verbose.
            self.covar = result.covar
            try:
                self.parmoptim_error = np.diag(self.covar)
            except:
                warnings.warn("PyCorrFit Warning: Error estimate not "+\
                              "possible, because we could not "+\
                              "calculate covariance matrix. Please "+\
                              "try reducing the number of fitting "+\
                              "parameters.")
                self.parmoptim_error = None
        else:
            self.parmoptim_error = None



def GetAlgorithmStringList():
    """
        Get supported fitting algorithms as strings.
        Returns two lists (that are key-sorted) for key and string.
    """
    A = Algorithms
    out1 = []
    out2 = []
    a = list(A.keys())
    a.sort()
    for key in a:
        out1.append(key)
        out2.append(A[key][1])
    return out1, out2
    

# As of version 0.8.3, we support several minimization methods for
# fitting data to experimental curves.
# These functions must be callable like scipy.optimize.leastsq. e.g.
# res = spopt.leastsq(self.fit_function, self.fitparms[:],
#                     args=(self.x), full_output=1)
Algorithms = dict()

# the original one is the least squares fit "leastsq"
Algorithms["Lev-Mar"] = ["leastsq", 
                         "Levenberg-Marquardt",
                         {"ftol" : 1.49012e-08,
                          "xtol" : 1.49012e-08,
                          }
                        ]

# simplex 
Algorithms["Nelder-Mead"] = ["nelder",
                             "Nelder-Mead (downhill simplex)",
                             {}
                             ]

# quasi-Newton method of Broyden, Fletcher, Goldfarb, and Shanno
Algorithms["BFGS"] = ["lbfgsb",
                      "BFGS (quasi-Newton)",
                      {}
                      ]

# modified Powell-method
Algorithms["Powell"] = ["powell",
                        "modified Powell (conjugate direction)",
                        {}
                        ]

# nonliner conjugate gradient method by Polak and Ribiere
Algorithms["SLSQP"] = ["slsqp",
                       "Sequential Linear Squares Programming",
                       {}
                       ]
